#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for saponification property package example.
Authors: Andrew Lee
"""

import pytest
from pyomo.environ import ConcreteModel, Constraint, Param, Suffix, value, Var
from pyomo.util.check_units import assert_units_consistent
from idaes.core import MaterialBalanceType, EnergyBalanceType, MaterialFlowBasis

from idaes.models.properties.examples.saponification_thermo import (
    SaponificationParameterBlock,
    SaponificationStateBlock,
    SaponificationPropertiesScaler,
)
from idaes.core.solvers import get_solver
from idaes.core.initialization import (
    BlockTriangularizationInitializer,
    InitializationStatus,
)


# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver("ipopt_v2")


class TestParamBlock(object):
    @pytest.fixture(scope="class")
    def model(self):
        model = ConcreteModel()
        model.params = SaponificationParameterBlock()

        return model

    @pytest.mark.unit
    def test_config(self, model):
        assert len(model.params.config) == 1

    @pytest.mark.unit
    def test_build(self, model):
        assert model.params.state_block_class is SaponificationStateBlock

        assert len(model.params.phase_list) == 1
        for i in model.params.phase_list:
            assert i == "Liq"

        assert len(model.params.component_list) == 5
        for i in model.params.component_list:
            assert i in ["H2O", "NaOH", "EthylAcetate", "SodiumAcetate", "Ethanol"]

        assert isinstance(model.params.cp_mol, Param)
        assert value(model.params.cp_mol) == 75.327

        assert isinstance(model.params.dens_mol, Param)
        assert value(model.params.dens_mol) == 55388

        assert isinstance(model.params.pressure_ref, Param)
        assert value(model.params.pressure_ref) == 101325

        assert isinstance(model.params.temperature_ref, Param)
        assert value(model.params.temperature_ref) == 298.15


class TestStateBlock(object):
    @pytest.fixture(scope="class")
    def model(self):
        model = ConcreteModel()
        model.params = SaponificationParameterBlock()

        model.props = model.params.build_state_block([1])

        return model

    @pytest.mark.unit
    def test_build(self, model):
        assert isinstance(model.props[1].flow_vol, Var)
        assert value(model.props[1].flow_vol) == 1

        assert isinstance(model.props[1].pressure, Var)
        assert value(model.props[1].pressure) == 101325

        assert isinstance(model.props[1].temperature, Var)
        assert value(model.props[1].temperature) == 298.15

        assert isinstance(model.props[1].conc_mol_comp, Var)
        assert len(model.props[1].conc_mol_comp) == 5
        for i in model.props[1].conc_mol_comp:
            assert value(model.props[1].conc_mol_comp[i]) == 100

        assert isinstance(model.props[1].conc_water_eqn, Constraint)
        assert len(model.props[1].flow_vol) == 1

        assert model.props[1].default_scaler is SaponificationPropertiesScaler

    @pytest.mark.unit
    def test_build_defined_state(self):
        model = ConcreteModel()
        model.params = SaponificationParameterBlock()

        model.props = model.params.build_state_block([1], defined_state=True)

        assert isinstance(model.props[1].flow_vol, Var)
        assert value(model.props[1].flow_vol) == 1

        assert isinstance(model.props[1].pressure, Var)
        assert value(model.props[1].pressure) == 101325

        assert isinstance(model.props[1].temperature, Var)
        assert value(model.props[1].temperature) == 298.15

        assert isinstance(model.props[1].conc_mol_comp, Var)
        assert len(model.props[1].conc_mol_comp) == 5
        for i in model.props[1].conc_mol_comp:
            assert value(model.props[1].conc_mol_comp[i]) == 100

        assert not hasattr(model.props[1], "conc_water_eqn")

    @pytest.mark.unit
    def test_get_material_flow_terms(self, model):
        for p in model.params.phase_list:
            for j in model.params.component_list:
                assert str(model.props[1].get_material_flow_terms(p, j)) == str(
                    model.props[1].flow_vol * model.props[1].conc_mol_comp[j]
                )

    @pytest.mark.unit
    def test_get_enthalpy_flow_terms(self, model):
        for p in model.params.phase_list:
            assert str(model.props[1].get_enthalpy_flow_terms(p)) == str(
                model.props[1].flow_vol
                * model.props[1].params.dens_mol
                * model.props[1].params.cp_mol
                * (model.props[1].temperature - model.props[1].params.temperature_ref)
            )

    @pytest.mark.unit
    def test_get_material_density_terms(self, model):
        for p in model.params.phase_list:
            for j in model.params.component_list:
                assert str(model.props[1].get_material_density_terms(p, j)) == str(
                    model.props[1].conc_mol_comp[j]
                )

    @pytest.mark.unit
    def test_get_energy_density_terms(self, model):
        for p in model.params.phase_list:
            assert str(model.props[1].get_energy_density_terms(p)) == str(
                model.props[1].params.dens_mol
                * model.props[1].params.cp_mol
                * (model.props[1].temperature - model.props[1].params.temperature_ref)
            )

    @pytest.mark.unit
    def test_default_material_balance_type(self, model):
        assert (
            model.props[1].default_material_balance_type()
            == MaterialBalanceType.componentPhase
        )

    @pytest.mark.unit
    def test_default_energy_balance_type(self, model):
        assert (
            model.props[1].default_energy_balance_type()
            == EnergyBalanceType.enthalpyTotal
        )

    @pytest.mark.unit
    def test_get_material_flow_basis(self, model):
        assert model.props[1].get_material_flow_basis() == MaterialFlowBasis.molar

    @pytest.mark.unit
    def test_define_state_vars(self, model):
        sv = model.props[1].define_state_vars()

        assert len(sv) == 4
        for i in sv:
            assert i in ["flow_vol", "conc_mol_comp", "temperature", "pressure"]

    @pytest.mark.unit
    def test_define_port_members(self, model):
        sv = model.props[1].define_state_vars()

        assert len(sv) == 4
        for i in sv:
            assert i in ["flow_vol", "conc_mol_comp", "temperature", "pressure"]

    @pytest.mark.unit
    def test_define_display_vars(self, model):
        sv = model.props[1].define_display_vars()

        assert len(sv) == 4
        for i in sv:
            assert i in [
                "Volumetric Flowrate",
                "Molar Concentration",
                "Temperature",
                "Pressure",
            ]

    @pytest.mark.unit
    def test_model_check_none(self, model, caplog):
        assert model.props[1].model_check() is None
        assert "Temperature set below lower bound" not in caplog.text
        assert "Temperature set above upper bound" not in caplog.text
        assert "Pressure set below lower bound" not in caplog.text
        assert "Pressure set above upper bound" not in caplog.text

    @pytest.mark.unit
    def test_model_check_low_T(self, model, caplog):
        model.props[1].temperature.value = 200
        assert model.props[1].model_check() is None
        assert "Temperature set below lower bound" in caplog.text
        assert "Temperature set above upper bound" not in caplog.text
        assert "Pressure set below lower bound" not in caplog.text
        assert "Pressure set above upper bound" not in caplog.text

    @pytest.mark.unit
    def test_model_check_high_T(self, model, caplog):
        model.props[1].temperature.value = 350
        assert model.props[1].model_check() is None
        assert "Temperature set below lower bound" not in caplog.text
        assert "Temperature set above upper bound" in caplog.text
        assert "Pressure set below lower bound" not in caplog.text
        assert "Pressure set above upper bound" not in caplog.text
        # Reset temperature
        model.props[1].temperature.value = 298.15

    @pytest.mark.unit
    def test_model_check_low_P(self, model, caplog):
        model.props[1].pressure.value = 1e2
        assert model.props[1].model_check() is None
        assert "Temperature set below lower bound" not in caplog.text
        assert "Temperature set above upper bound" not in caplog.text
        assert "Pressure set below lower bound" in caplog.text
        assert "Pressure set above upper bound" not in caplog.text

    @pytest.mark.unit
    def test_model_check_high_P(self, model, caplog):
        model.props[1].pressure.value = 1e7
        assert model.props[1].model_check() is None
        assert "Temperature set below lower bound" not in caplog.text
        assert "Temperature set above upper bound" not in caplog.text
        assert "Pressure set below lower bound" not in caplog.text
        assert "Pressure set above upper bound" in caplog.text
        # Reset pressure
        model.props[1].pressure.value = 101325

    @pytest.mark.unit
    def test_initialize(self, model):
        assert not model.props[1].flow_vol.fixed
        assert not model.props[1].temperature.fixed
        assert not model.props[1].pressure.fixed
        for i in model.props[1].conc_mol_comp:
            assert not model.props[1].conc_mol_comp[i].fixed

        model.props.initialize(hold_state=False, outlvl=1)

        assert not model.props[1].flow_vol.fixed
        assert not model.props[1].temperature.fixed
        assert not model.props[1].pressure.fixed
        for i in model.props[1].conc_mol_comp:
            assert not model.props[1].conc_mol_comp[i].fixed

    @pytest.mark.unit
    def test_initialize_hold(self, model):
        assert not model.props[1].flow_vol.fixed
        assert not model.props[1].temperature.fixed
        assert not model.props[1].pressure.fixed
        for i in model.props[1].conc_mol_comp:
            assert not model.props[1].conc_mol_comp[i].fixed

        flags = model.props.initialize(hold_state=True)

        assert model.props[1].flow_vol.fixed
        assert model.props[1].temperature.fixed
        assert model.props[1].pressure.fixed
        for i in model.props[1].conc_mol_comp:
            assert model.props[1].conc_mol_comp[i].fixed

        model.props.release_state(flags, outlvl=1)

        assert not model.props[1].flow_vol.fixed
        assert not model.props[1].temperature.fixed
        assert not model.props[1].pressure.fixed
        for i in model.props[1].conc_mol_comp:
            assert not model.props[1].conc_mol_comp[i].fixed

    @pytest.mark.component
    def check_units(self, model):
        assert_units_consistent(model)


@pytest.mark.component
def test_initializer():
    model = ConcreteModel()
    model.params = SaponificationParameterBlock()

    model.props = model.params.build_state_block([1])

    assert model.props.default_initializer is BlockTriangularizationInitializer

    initializer = model.props.default_initializer()
    initializer.initialize(model.props)

    assert initializer.summary[model.props]["status"] == InitializationStatus.Ok


class TestSaponificationPropertiesScaler:
    @pytest.mark.unit
    def test_variable_scaling_routine(self):
        model = ConcreteModel()
        model.params = SaponificationParameterBlock()

        model.props = model.params.build_state_block([1], defined_state=False)

        scaler = model.props[1].default_scaler()
        assert isinstance(scaler, SaponificationPropertiesScaler)

        scaler.variable_scaling_routine(model.props[1])

        assert isinstance(model.props[1].scaling_factor, Suffix)

        sfx = model.props[1].scaling_factor
        assert len(sfx) == 8
        assert sfx[model.props[1].flow_vol] == pytest.approx(1e2, rel=1e-8)
        assert sfx[model.props[1].pressure] == pytest.approx(1e-5, rel=1e-8)
        assert sfx[model.props[1].temperature] == pytest.approx(1 / 310.65, rel=1e-8)
        for k, v in model.props[1].conc_mol_comp.items():
            if k == "H2O":
                assert sfx[v] == pytest.approx(1e-4, rel=1e-8)
            else:
                assert sfx[v] == pytest.approx(1e-2, rel=1e-8)

    @pytest.mark.unit
    def test_constraint_scaling_routine(self):
        model = ConcreteModel()
        model.params = SaponificationParameterBlock()

        model.props = model.params.build_state_block([1], defined_state=False)

        scaler = model.props[1].default_scaler()
        assert isinstance(scaler, SaponificationPropertiesScaler)

        scaler.constraint_scaling_routine(model.props[1])

        assert isinstance(model.props[1].scaling_factor, Suffix)

        sfx = model.props[1].scaling_factor
        assert len(sfx) == 1
        assert sfx[model.props[1].conc_water_eqn] == pytest.approx(1e-4, rel=1e-8)

    @pytest.mark.unit
    def test_constraint_scaling_routine_defined_state(self):
        model = ConcreteModel()
        model.params = SaponificationParameterBlock()

        model.props = model.params.build_state_block([1], defined_state=True)

        scaler = model.props[1].default_scaler()
        assert isinstance(scaler, SaponificationPropertiesScaler)

        scaler.constraint_scaling_routine(model.props[1])

        # No constraints, so there should be no Suffix
        assert not hasattr(model.props[1], "scaling_factor")

    @pytest.mark.unit
    def test_scale_model(self):
        model = ConcreteModel()
        model.params = SaponificationParameterBlock()

        model.props = model.params.build_state_block([1], defined_state=False)

        scaler = model.props[1].default_scaler()
        assert isinstance(scaler, SaponificationPropertiesScaler)

        scaler.scale_model(model.props[1])

        assert isinstance(model.props[1].scaling_factor, Suffix)

        sfx = model.props[1].scaling_factor
        assert len(sfx) == 9
        assert sfx[model.props[1].flow_vol] == pytest.approx(1e2, rel=1e-8)
        assert sfx[model.props[1].pressure] == pytest.approx(1e-5, rel=1e-8)
        assert sfx[model.props[1].temperature] == pytest.approx(1 / 310.65, rel=1e-8)
        for k, v in model.props[1].conc_mol_comp.items():
            if k == "H2O":
                assert sfx[v] == pytest.approx(1e-4, rel=1e-8)
            else:
                assert sfx[v] == pytest.approx(1e-2, rel=1e-8)
        assert sfx[model.props[1].conc_water_eqn] == pytest.approx(1e-4, rel=1e-8)



#!/usr/bin/env python
"""
Institute for the Design of Advanced Energy Systems
"""
from pathlib import Path
import os
import sys
from setuptools import setup, find_namespace_packages
from typing import List, Tuple


def get_version():
    code_file = os.path.join("idaes", "ver.py")
    code = open(code_file).read()
    local_namespace = {}
    exec(code, {}, local_namespace)
    return local_namespace["__version__"]


NAME = "idaes-pse"
VERSION = get_version()
README = open("README.md").read()
README = README[README.find("#") :]  # ignore everything before title


class ExtraDependencies:
    """
    A convenience shorthand to define and combine dependencies for ``extras_require``.

    >>> extras = ExtraDependencies()
    >>> extras["testing"]
    ['pytest', 'addheader', 'pyyaml']
    >>> list(extras)
    ['ui', 'omlt', 'grid', 'coolprop', 'testing']
    >>> dict(extras)
    {'testing': ['pytest', 'addheader', 'pyyaml'], 'omlt': ['omlt==1.1', 'tensorflow', ...], ...}
    """

    ui = [
        # FIXME this must be changed to the PyPI distribution for the release
        "idaes-ui",
    ]
    omlt = [
        "omlt==1.1",  # fix the version for now as package evolves
        "tensorflow",
        "onnx",
    ]
    grid = [
        "gridx-prescient>=2.2.1",  # idaes.tests.prescient
    ]
    coolprop = [
        "coolprop",  # idaes.generic_models.properties.general.coolprop
    ]
    testing = [
        "pytest",
        "addheader",
        "pyyaml",
    ]

    def __init__(self):
        self._data = dict(type(self).__dict__)

    def keys(self):
        for name, attr in self._data.items():
            if not name.startswith("_") and isinstance(attr, list):
                yield name

    def __getitem__(self, key):
        return self._data[key]


kwargs = dict(
    zip_safe=False,
    name=NAME,
    version=VERSION,
    packages=find_namespace_packages(
        include=[
            "idaes*",
        ]
    ),
    # Put abstract (non-versioned) deps here.
    # Concrete dependencies go in requirements[-dev].txt
    install_requires=[
        "pyomo >= 6.8.2",
        "pint >= 0.24.1",  # required to use Pyomo units. Pint 0.24.1 needed for Python 3.9 support
        "networkx",  # required to use Pyomo network
        "numpy>=1,<3",
        # pandas constraint added on 2023-08-30 b/c bug in v2.1
        # see IDAES/idaes-pse#1253
        "pandas!=2.1.0,<3",
        "scipy",
        "sympy",  # idaes.core.util.expr_doc
        "matplotlib",
        "click>=8",
    ],
    entry_points={
        "console_scripts": [
            "idaes = idaes.commands.base:command_base",
        ]
    },
    # Only installed if [<key>] is added to package name
    extras_require=dict(ExtraDependencies()),
    package_data={
        # If any package contains these files, include them:
        "": [
            "*.template",
            "*.json",
            "*.yaml",
            "*.svg",
            "*.png",
            "*.jpg",
            "*.csv",
            "*.ipynb",
            "*.txt",
            "*.js",
            "*.css",
            "*.html",
            "*.json.gz",
            "*.dat",
            "*.h5",
            "*.pb",  # for Keras Surrogate folder
            "*.data-00000-of-00001",  # for Keras Surrogate folder
            "*.index",  # for Keras Surrogate folder
            "*.trc",
            "*.nl",
            "*.keras",  # idaes/core/surrogate/tests/data/keras_models
            "*.onnx",
        ]
    },
    include_package_data=True,
    maintainer="Keith Beattie",
    maintainer_email="ksbeattie@lbl.gov",
    url="https://idaes.org",
    license="BSD ",
    platforms=["any"],
    description="IDAES Process Systems Engineering Framework",
    long_description=README,
    long_description_content_type="text/markdown",
    keywords=[NAME, "energy systems", "chemical engineering", "process modeling"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: Unix",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering :: Chemistry",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)

setup(**kwargs)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################

__author__ = "Douglas Allan"

import pytest
import numpy as np

import pyomo.environ as pyo
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from idaes.core.util.constants import Constants
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.exceptions import ConfigurationError
import idaes.models_extra.power_generation.unit_models.soc_submodels as soc
import idaes.models_extra.power_generation.unit_models.soc_submodels.common as common
import idaes.models_extra.power_generation.unit_models.soc_submodels.testing as soc_testing
from idaes.core.solvers import get_solver

solver = get_solver("ipopt")


def common_components(nt, nz, ncomp, nreact):
    return {
        pyo.Var: {
            "temperature_z": nz * nt,
            "temperature_deviation_x": nz * nt,
            "conc_mol_comp_ref": nz * nt * ncomp,
            "conc_mol_comp_deviation_x": nz * nt * ncomp,
            "material_flux_x": nz * nt * ncomp,
            "heat_flux_x0": nz * nt,
            "current_density": nz * nt,
            "length_z": 1,
            "length_y": 1,
            "heat_flux_x1": nz * nt,
            "mole_frac_comp": nz * nt * ncomp,
            "log_mole_frac_comp": nz * nt * nreact,
            "activation_potential": nz * nt,
            "activation_potential_alpha1": 1,
            "activation_potential_alpha2": 1,
            "exchange_current_exponent_comp": nreact,
            "exchange_current_log_preexponential_factor": 1,
            "exchange_current_activation_energy": 1,
        },
        pyo.Constraint: {
            "mole_frac_comp_eqn": nz * nt * ncomp,
            "log_mole_frac_comp_eqn": nz * nt * nreact,
            "material_flux_x_eqn": nz * nt * ncomp,
            "activation_potential_eqn": nz * nt,
            "heat_flux_x_eqn": nz * nt,
        },
        pyo.Expression: {
            "temperature": nz * nt,
            "conc_mol_comp": nz * nt * ncomp,
            "pressure": nz * nt,
            "ds_rxn": nz * nt,
            "dh_rxn": nz * nt,
            "dg_rxn": nz * nt,
            "potential_nernst": nz * nt,
            "log_exchange_current_density": nz * nt,
            "reaction_rate_per_unit_area": nz * nt,
            "voltage_drop_total": nz * nt,
        },
    }


@pytest.fixture
def modelFuel():
    time_set = [0]
    zfaces = np.linspace(0, 1, 6).tolist()
    fuel_comps = ["H2", "N2", "H2O"]
    m = soc_testing._cell_flowsheet_model(
        dynamic=False, time_set=time_set, zfaces=zfaces
    )
    iznodes = m.fs.iznodes
    # time_units = m.fs.time_units
    tset = m.fs.config.time
    comps = m.fs.comps = pyo.Set(initialize=fuel_comps)

    m.fs.temperature_deviation_x = pyo.Var(
        tset, iznodes, initialize=0, units=pyo.units.K
    )
    m.fs.conc_mol_comp_ref = pyo.Var(
        tset, iznodes, comps, initialize=1, units=pyo.units.mol / pyo.units.m**3
    )
    m.fs.conc_mol_comp_deviation_x = pyo.Var(
        tset, iznodes, comps, initialize=0, units=pyo.units.mol / pyo.units.m**3
    )
    m.fs.material_flux_x = pyo.Var(
        tset,
        iznodes,
        comps,
        initialize=0,
        units=pyo.units.mol / (pyo.units.s * pyo.units.m**2),
    )
    m.fs.heat_flux_x0 = pyo.Var(
        tset, iznodes, initialize=0, units=pyo.units.W / pyo.units.m**2
    )
    m.fs.fuel_triple_phase_boundary = soc.SocTriplePhaseBoundary(
        control_volume_zfaces=zfaces,
        length_z=m.fs.length_z,
        length_y=m.fs.length_y,
        component_list=fuel_comps,
        reaction_stoichiometry={"H2": -0.5, "H2O": 0.5, "N2": 0, "e^-": 1.0},
        inert_species=["N2"],
        below_electrolyte=True,
        current_density=m.fs.current_density,
        temperature_z=m.fs.temperature_z,
        temperature_deviation_x=m.fs.temperature_deviation_x,
        heat_flux_x0=m.fs.heat_flux_x0,
        conc_mol_comp_ref=m.fs.conc_mol_comp_ref,
        conc_mol_comp_deviation_x=m.fs.conc_mol_comp_deviation_x,
        material_flux_x=m.fs.material_flux_x,
    )
    m.fs.temperature_deviation_x.fix(0)
    m.fs.heat_flux_x0.fix(0)
    m.fs.conc_mol_comp_ref.fix(1)
    m.fs.conc_mol_comp_deviation_x.fix(0)

    m.fs.fuel_triple_phase_boundary.exchange_current_log_preexponential_factor.fix(
        pyo.log(1.375e10)
    )
    m.fs.fuel_triple_phase_boundary.exchange_current_activation_energy.fix(120e3)
    m.fs.fuel_triple_phase_boundary.activation_potential_alpha1.fix(0.5)
    m.fs.fuel_triple_phase_boundary.activation_potential_alpha2.fix(0.5)

    m.fs.fuel_triple_phase_boundary.exchange_current_exponent_comp["H2"].fix(1)
    m.fs.fuel_triple_phase_boundary.exchange_current_exponent_comp["H2O"].fix(1)

    return m


@pytest.fixture
def modelOxygen():
    time_set = [0, 1, 2]
    zfaces = np.linspace(0, 1, 8).tolist()
    o2_comps = ["O2", "N2"]
    m = pyo.ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False, time_set=time_set, time_units=pyo.units.s)
    m.fs.oxygen_triple_phase_boundary = soc.SocTriplePhaseBoundary(
        control_volume_zfaces=zfaces,
        component_list=o2_comps,
        reaction_stoichiometry={"O2": -0.25, "e^-": -1.0},
        inert_species=["N2"],
    )
    m.fs.oxygen_triple_phase_boundary.temperature_z.fix(0)
    m.fs.oxygen_triple_phase_boundary.current_density.fix(0)

    m.fs.oxygen_triple_phase_boundary.temperature_deviation_x.fix(0)
    m.fs.oxygen_triple_phase_boundary.heat_flux_x0.fix(0)
    m.fs.oxygen_triple_phase_boundary.conc_mol_comp_ref.fix(1)
    m.fs.oxygen_triple_phase_boundary.conc_mol_comp_deviation_x.fix(0)

    m.fs.oxygen_triple_phase_boundary.exchange_current_log_preexponential_factor.fix(
        pyo.log(1.375e10)
    )
    m.fs.oxygen_triple_phase_boundary.exchange_current_activation_energy.fix(120e3)
    m.fs.oxygen_triple_phase_boundary.activation_potential_alpha1.fix(0.5)
    m.fs.oxygen_triple_phase_boundary.activation_potential_alpha2.fix(0.5)

    m.fs.oxygen_triple_phase_boundary.exchange_current_exponent_comp["O2"].fix(0.25)

    return m


@pytest.mark.build
@pytest.mark.unit
def test_build_fuel(modelFuel):
    tpb = modelFuel.fs.fuel_triple_phase_boundary
    nz = len(modelFuel.fs.iznodes)
    nt = len(modelFuel.fs.time)
    ncomp = len(tpb.component_list)
    nreact = 2
    soc_testing._build_test_utility(
        block=tpb,
        comp_dict=common_components(nt, nz, ncomp, nreact),
        references=[
            "temperature_z",
            "temperature_deviation_x",
            "conc_mol_comp_ref",
            "conc_mol_comp_deviation_x",
            "material_flux_x",
            "heat_flux_x0",
            "current_density",
            "length_z",
            "length_y",
        ],
    )
    assert degrees_of_freedom(tpb) == 0


@pytest.mark.component
def test_units_model_fuel(modelFuel):
    assert_units_consistent(modelFuel)


@pytest.mark.build
@pytest.mark.unit
def test_build_oxygen(modelOxygen):
    tpb = modelOxygen.fs.oxygen_triple_phase_boundary
    nz = len(tpb.iznodes)
    nt = len(modelOxygen.fs.time)
    ncomp = len(tpb.component_list)
    nreact = 1
    soc_testing._build_test_utility(
        block=tpb,
        comp_dict=common_components(nt, nz, ncomp, nreact),
    )

    assert degrees_of_freedom(tpb) == 0


@pytest.mark.component
def test_units_model_oxygen(modelOxygen):
    assert_units_consistent(modelOxygen)


@pytest.mark.solver
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_initialization_fuel(modelFuel):
    modelFuel.fs.fuel_triple_phase_boundary.initialize(
        fix_x0=True, optarg={"nlp_scaling_method": "user-scaling"}
    )


@pytest.mark.unit
def test_extra_inert():
    time_set = [0, 1, 2]
    zfaces = np.linspace(0, 1, 8).tolist()
    o2_comps = ["O2", "N2"]
    m = pyo.ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False, time_set=time_set, time_units=pyo.units.s)
    with pytest.raises(
        ConfigurationError,
        match="fs.oxygen_triple_phase_boundary invalid component in inert_species "
        "argument. H2O is not in the provided component list.",
    ):
        m.fs.oxygen_triple_phase_boundary = soc.SocTriplePhaseBoundary(
            control_volume_zfaces=zfaces,
            component_list=o2_comps,
            reaction_stoichiometry={"O2": -0.25, "Vac": -0.5, "O^2-": 0.5, "e^-": -1.0},
            inert_species=["N2", "H2O"],
        )


def modelFuelAndOxygen(include_solid_species):
    time_set = [0]
    zfaces = [0, 1]
    fuel_comps = ["H2", "N2", "H2O"]
    oxygen_comps = ["O2", "N2"]
    fuel_stoich = {
        "H2": -1.0,
        "H2O": 1.0,
        "N2": 0.0,
        "e^-": 2.0,
    }
    oxygen_stoich = {"O2": -1.0, "e^-": -4.0}
    if include_solid_species:
        fuel_stoich["Vac"] = 1.0
        fuel_stoich["O^2-"] = -1.0
        oxygen_stoich["Vac"] = -2.0
        oxygen_stoich["O^2-"] = 2.0

    m = soc_testing._cell_flowsheet_model(
        dynamic=False, time_set=time_set, zfaces=zfaces
    )
    iznodes = m.fs.iznodes
    # time_units = m.fs.time_units
    tset = m.fs.config.time
    comps = m.fs.comps = pyo.Set(initialize=fuel_comps)

    m.fs.temperature_deviation_x = pyo.Var(
        tset, iznodes, initialize=0, units=pyo.units.K
    )
    m.fs.fuel_triple_phase_boundary = soc.SocTriplePhaseBoundary(
        control_volume_zfaces=zfaces,
        length_z=m.fs.length_z,
        length_y=m.fs.length_y,
        component_list=fuel_comps,
        reaction_stoichiometry=fuel_stoich,
        inert_species=["N2"],
        below_electrolyte=True,
        current_density=m.fs.current_density,
        temperature_z=m.fs.temperature_z,
        temperature_deviation_x=m.fs.temperature_deviation_x,
    )
    m.fs.oxygen_triple_phase_boundary = soc.SocTriplePhaseBoundary(
        control_volume_zfaces=zfaces,
        length_z=m.fs.length_z,
        length_y=m.fs.length_y,
        component_list=oxygen_comps,
        reaction_stoichiometry=oxygen_stoich,
        inert_species=["N2"],
        below_electrolyte=False,
        current_density=m.fs.current_density,
        temperature_z=m.fs.temperature_z,
        temperature_deviation_x=m.fs.temperature_deviation_x,
        heat_flux_x0=m.fs.fuel_triple_phase_boundary.heat_flux_x1,
    )

    m.fs.fuel_triple_phase_boundary.exchange_current_log_preexponential_factor.fix(
        pyo.log(1.375e10)
    )
    m.fs.fuel_triple_phase_boundary.exchange_current_activation_energy.fix(120e3)
    m.fs.fuel_triple_phase_boundary.activation_potential_alpha1.fix(0.5)
    m.fs.fuel_triple_phase_boundary.activation_potential_alpha2.fix(0.5)

    m.fs.fuel_triple_phase_boundary.exchange_current_exponent_comp["H2"].fix(1)
    m.fs.fuel_triple_phase_boundary.exchange_current_exponent_comp["H2O"].fix(1)

    m.fs.oxygen_triple_phase_boundary.exchange_current_log_preexponential_factor.fix(
        pyo.log(1.375e10)
    )
    m.fs.oxygen_triple_phase_boundary.exchange_current_activation_energy.fix(120e3)
    m.fs.oxygen_triple_phase_boundary.activation_potential_alpha1.fix(0.5)
    m.fs.oxygen_triple_phase_boundary.activation_potential_alpha2.fix(0.5)

    T = 1000
    m.fs.temperature_z.fix(T)
    m.fs.temperature_deviation_x.fix(0)
    m.fs.fuel_triple_phase_boundary.conc_mol_comp_deviation_x.fix(0)
    m.fs.oxygen_triple_phase_boundary.conc_mol_comp_deviation_x.fix(0)

    m.fs.oxygen_triple_phase_boundary.exchange_current_exponent_comp["O2"].fix(0.25)

    C_tot = 1.2e5 / pyo.value(Constants.gas_constant * T)

    m.fs.fuel_triple_phase_boundary.conc_mol_comp_ref[0, :, "H2O"].fix(0.5 * C_tot)
    m.fs.fuel_triple_phase_boundary.conc_mol_comp_ref[0, :, "H2"].fix(0.05 * C_tot)
    m.fs.fuel_triple_phase_boundary.conc_mol_comp_ref[0, :, "N2"].fix(0.45 * C_tot)

    m.fs.oxygen_triple_phase_boundary.conc_mol_comp_ref[0, :, "O2"].fix(0.21 * C_tot)
    m.fs.oxygen_triple_phase_boundary.conc_mol_comp_ref[0, :, "N2"].fix(0.79 * C_tot)

    m.fs.fuel_triple_phase_boundary.heat_flux_x1.fix(0)
    m.fs.current_density.fix(1000)

    @m.fs.Expression(m.fs.time, m.fs.fuel_triple_phase_boundary.iznodes)
    def net_energy_flux_out(b, t, iz):
        return (
            -sum(
                b.fuel_triple_phase_boundary.material_flux_x[t, iz, comp]
                * common._comp_enthalpy_expr(
                    b.fuel_triple_phase_boundary.temperature[t, iz], comp
                )
                for comp in b.fuel_triple_phase_boundary.component_list
            )
            - b.fuel_triple_phase_boundary.heat_flux_x0[t, iz]
            + sum(
                b.oxygen_triple_phase_boundary.material_flux_x[t, iz, comp]
                * common._comp_enthalpy_expr(
                    b.oxygen_triple_phase_boundary.temperature[t, iz], comp
                )
                for comp in b.oxygen_triple_phase_boundary.component_list
            )
            + b.oxygen_triple_phase_boundary.heat_flux_x1[t, iz]
        )

    @m.fs.Expression(m.fs.time, m.fs.fuel_triple_phase_boundary.iznodes)
    def voltage_difference(b, t, iz):
        return (
            b.fuel_triple_phase_boundary.potential_nernst[t, iz]
            + b.oxygen_triple_phase_boundary.potential_nernst[t, iz]
            - (
                b.fuel_triple_phase_boundary.voltage_drop_total[t, iz]
                + b.oxygen_triple_phase_boundary.voltage_drop_total[t, iz]
            )
        )

    @m.fs.Expression(m.fs.time, m.fs.fuel_triple_phase_boundary.iznodes)
    def electric_work_flux(b, t, iz):
        return b.voltage_difference[t, iz] * b.current_density[t, iz]

    @m.fs.Expression(m.fs.time, m.fs.fuel_triple_phase_boundary.iznodes)
    def energy_created(b, t, iz):
        return b.net_energy_flux_out[t, iz] + b.electric_work_flux[t, iz]

    m.fs.element_list = pyo.Set(initialize=["H", "O", "N"], ordered=True)

    @m.fs.Expression(
        m.fs.time, m.fs.fuel_triple_phase_boundary.iznodes, m.fs.element_list
    )
    def element_created(b, t, iz, i):
        return -sum(
            common._element_dict[i][j]
            * b.fuel_triple_phase_boundary.material_flux_x[t, iz, j]
            for j in b.fuel_triple_phase_boundary.component_list
        ) + sum(
            common._element_dict[i][j]
            * b.oxygen_triple_phase_boundary.material_flux_x[t, iz, j]
            for j in b.oxygen_triple_phase_boundary.component_list
        )

    return m


def conservation_tester(m):
    for P_fuel in np.linspace(1e5, 5e5, 3):
        for T in np.linspace(900, 1100, 4):
            m.fs.temperature_z.fix(T)
            C_tot = P_fuel / pyo.value(Constants.gas_constant * T)
            for y_H2 in np.linspace(0.1, 0.5, 3):
                m.fs.fuel_triple_phase_boundary.conc_mol_comp_ref[0, :, "H2"].fix(
                    y_H2 * C_tot
                )
                m.fs.fuel_triple_phase_boundary.conc_mol_comp_ref[0, :, "N2"].fix(
                    0.4 * C_tot
                )
                m.fs.fuel_triple_phase_boundary.conc_mol_comp_ref[0, :, "H2O"].fix(
                    (0.6 - y_H2) * C_tot
                )
                for y_O2 in np.linspace(0.1, 0.3, 3):
                    m.fs.oxygen_triple_phase_boundary.conc_mol_comp_ref[0, :, "O2"].fix(
                        y_O2 * C_tot
                    )
                    m.fs.oxygen_triple_phase_boundary.conc_mol_comp_ref[0, :, "N2"].fix(
                        (1 - y_O2) * C_tot
                    )
                    for J in np.linspace(-1000, 1000, 5):
                        m.fs.current_density.fix(J)
                        solver.solve(m)
                        assert pyo.value(m.fs.energy_created[0, 1]) == pytest.approx(
                            0, rel=1e-4
                        )
                        for i in m.fs.element_list:
                            assert pyo.value(
                                m.fs.element_created[0, 1, i]
                            ) == pytest.approx(0, rel=1e-4)


@pytest.mark.solver
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.integration
def test_conservation_no_solid_species():
    m = modelFuelAndOxygen(False)
    solver.solve(m)
    conservation_tester(m)


@pytest.mark.solver
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.integration
def test_conservation_solid_species():
    m = modelFuelAndOxygen(True)
    solver.solve(m)
    conservation_tester(m)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################

from importlib import resources
from numbers import Number
from pathlib import Path
from typing import Dict, Union

import pytest
import pandas as pd


# define custom type for type hinting
PrescientOptions = Dict[str, Union[str, bool, Number, dict]]


class Test5Bus:
    "Simple test using 5bus use case"

    @pytest.fixture
    def data_path(self) -> Path:
        # NOTE here we want the path to the entire 5bus directory
        # we need to specify __init__.py as a workaround for Python 3.9,
        # where importlib.resources.path() requires the resource to be a file
        # directories are not supported and will raise an error if attempted
        with resources.path("idaes.tests.prescient.5bus", "__init__.py") as pkg_file:
            return Path(pkg_file).parent

    @pytest.mark.unit
    def test_data_path_available(self, data_path: Path):
        assert data_path.is_dir()

    @pytest.fixture
    def output_dir(self, tmp_path: Path) -> Path:
        path = tmp_path / "5bus_output"
        path.mkdir()
        return path

    @pytest.fixture
    def prescient_options(self, data_path: Path, output_dir: Path) -> PrescientOptions:
        return {
            "data_path": str(data_path),
            "input_format": "rts-gmlc",
            "simulate_out_of_sample": True,
            "run_sced_with_persistent_forecast_errors": True,
            "output_directory": str(output_dir),
            "start_date": "07-10-2020",
            "num_days": 2,
            "sced_horizon": 1,
            "ruc_mipgap": 0.01,
            "reserve_factor": 0.1,
            "deterministic_ruc_solver": "cbc",
            "deterministic_ruc_solver_options": {
                "feas": "off",
                "DivingF": "on",
            },
            "sced_solver": "cbc",
            "sced_frequency_minutes": 60,
            "ruc_horizon": 36,
            "compute_market_settlements": True,
            "monitor_all_contingencies": False,
            "output_solver_logs": False,
            "price_threshold": 1000,
            "contingency_price_threshold": 100,
            "reserve_price_threshold": 5,
        }

    @pytest.fixture
    def run_simulator(self, prescient_options: PrescientOptions) -> None:
        prescient_simulator = pytest.importorskip(
            "prescient.simulator",
            reason="Prescient (optional dependency) not available",
        )

        sim = prescient_simulator.Prescient()
        sim.simulate(**prescient_options)

    @pytest.fixture
    def simulation_results_table(
        self,
        run_simulator,
        output_dir: Path,
        name: str = "overall_simulation_output.csv",
    ) -> pd.DataFrame:

        path = output_dir / name
        return pd.read_csv(path)

    @pytest.mark.component
    # TODO use a more specific test to validate simulation output
    def test_simulation_results(self, simulation_results_table: pd.DataFrame):
        assert not simulation_results_table.empty



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""Get info about the environment IDAES is running in and the IDAES version
"""
# TODO: Missing docstrings
# pylint: disable=missing-function-docstring

__author__ = "John Eslick"

import click

from idaes.commands import cb
from idaes.core.util.env_info import EnvironmentInfo


@cb.command(
    name="environment-info", help="Print information about idaes, OS, dependencies..."
)
@click.option("--solver", multiple=True, help="Add solvers to list of solvers to check")
@click.option("--json", default=None, help="Write output to a file")
def environment_info(solver, json):
    info = EnvironmentInfo(additional_solvers=solver)
    d = info.to_dict()
    if json is None:
        for k, v in d.items():
            click.echo("")
            click.echo(f"{k}")
            for l, q in v.items():
                click.echo(f"    {l}: {q}")
    else:
        info.to_json(json)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Unit model to scale SOC variables to and from cell-scale to module-scale and
translate between the FTPx variables assumed at the cell level to a general
property package. The only module-level variable, besides those contained
in the StateBlocks, is ``number_cells``.

A config block for the SOC unit model must be provided. Further documentation
is provided in the cell model file.

Manipulated variables:
    - ``potential_cell[t]``: Voltage across single cell

Ports:
    * ``fuel_inlet``
    * ``fuel_outlet``
    * ``oxygen_inlet``
    * ``oxygen_outlet``

Instances of ``Var`` that must be fixed:
    - ``number_cells``: Number of cells in module

Expressions:
    - ``electrical_work[t]``: Rate of energy added to module. Greater than zero means energy added to module
      (electrolysis mode) and less than zero means energy removed from module (fuel cell mode)
"""
# TODO: Missing docstrings
# pylint: disable=missing-class-docstring

# TODO: Look into protected access issues
# pylint: disable=protected-access

__author__ = "Douglas Allan"


from functools import partial

from pyomo.common.config import ConfigValue, ConfigBlock
import pyomo.environ as pyo

from idaes.core import declare_process_block_class, UnitModelBlockData
from idaes.core.util.config import is_physical_parameter_block
import idaes.models_extra.power_generation.unit_models.soc_submodels as soc
import idaes.models_extra.power_generation.unit_models.soc_submodels.common as common
import idaes.core.util.scaling as iscale
from idaes.core.solvers import get_solver

import idaes.logger as idaeslog


@declare_process_block_class("SolidOxideModuleSimple")
class SolidOxideModuleSimpleData(UnitModelBlockData):
    CONFIG = UnitModelBlockData.CONFIG()
    # CONFIG = ConfigBlock()
    CONFIG.declare(
        "solid_oxide_cell_config",
        ConfigBlock(
            implicit=True,
            description="Config block for the SolidOxideCell",
        ),
    )
    CONFIG.declare(
        "fuel_property_package",
        ConfigValue(
            default=None,
            domain=is_physical_parameter_block,
            description="Property package to use for fuel-side stream",
            doc="""Property parameter object used to define property
                    calculations for the fuel-side stream,
                    **default** - None.
                    **Valid values:** {
                    **PhysicalParameterObject** - a PhysicalParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "fuel_property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property package "
            "of the fuel-side stream",
            doc="""A ConfigBlock with arguments to be passed to the property
                    block associated with the fuel-side stream,
                    **default** - None.
                    **Valid values:** {
                    see property package for documentation.}""",
        ),
    )
    CONFIG.declare(
        "oxygen_property_package",
        ConfigValue(
            default=None,
            domain=is_physical_parameter_block,
            description="Property package to use for oxygen-side stream",
            doc="""Property parameter object used to define property
                        calculations for the oxygen-side stream,
                        **default** - None.
                        **Valid values:** {
                        **PhysicalParameterObject** - a PhysicalParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "oxygen_property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property package "
            "of the oxygen-side stream",
            doc="""A ConfigBlock with arguments to be passed to the property
                        block associated with the oxygen-side stream,
                        **default** - None.
                        **Valid values:** {
                        see property package for documentation.}""",
        ),
    )

    def build(self):
        super().build()
        tset = self.flowsheet().config.time

        self.number_cells = pyo.Var(
            initialize=1e5,
            units=pyo.units.dimensionless,
            doc="Number of cells in SOC module",
        )

        self.solid_oxide_cell = soc.SolidOxideCell(
            **self.config.solid_oxide_cell_config
        )

        def rule_absent_comp(blk, t, j, props):
            return props[t].flow_mol_comp[j] == 1e-20

        def rule_temperature(blk, t, props, port):
            return props[t].temperature == port.temperature[t]

        def rule_pressure(blk, t, props, port):
            return props[t].pressure == port.pressure[t]

        def rule_flow_mol_comp(blk, t, j, props, port):
            return (
                props[t].flow_mol_comp[j]
                == blk.number_cells * port.flow_mol[t] * port.mole_frac_comp[t, j]
            )

        def rule_mole_frac(blk, t, port, comps):
            return sum(port.mole_frac_comp[t, j] for j in comps) == 1

        for side in ["fuel", "oxygen"]:
            param_block = getattr(self.config, f"{side}_property_package")
            package_args = getattr(self.config, f"{side}_property_package_args")
            setattr(
                self,
                f"{side}_properties_in",
                param_block.build_state_block(
                    self.flowsheet().time,
                    doc=f"Material properties in {side}-side inlet stream",
                    defined_state=True,
                    has_phase_equilibrium=False,
                    **package_args,
                ),
            )
            setattr(
                self,
                f"{side}_properties_out",
                param_block.build_state_block(
                    self.flowsheet().time,
                    doc=f"Material properties in {side}-side outlet stream",
                    defined_state=False,
                    has_phase_equilibrium=False,
                    **package_args,
                ),
            )
            side_comps = getattr(self.solid_oxide_cell, f"{side}_component_list")
            # Support absent components in property package to support legacy property packages that do not
            # permit users to exclude components.
            absent_comp_list = common._set_and_get_attr(
                self,
                f"absent_{side}_component_list",
                pyo.Set(
                    initialize=param_block.component_list - side_comps,
                    ordered=True,
                    doc=f"Components in the {side}-side property package that are not actually present in the stream.",
                ),
            )

            for direction in ["in", "out"]:
                props = getattr(self, f"{side}_properties_{direction}")
                port_name = f"{side}_{direction}let"  # e.g., fuel_inlet, oxygen_outlet
                port = getattr(self.solid_oxide_cell, port_name)
                setattr(
                    self,
                    f"{port_name}_temperature_eqn",
                    pyo.Constraint(
                        tset, rule=partial(rule_temperature, props=props, port=port)
                    ),
                )
                setattr(
                    self,
                    f"{port_name}_pressure_eqn",
                    pyo.Constraint(
                        tset, rule=partial(rule_pressure, props=props, port=port)
                    ),
                )
                setattr(
                    self,
                    f"{port_name}_flow_mol_comp_eqn",
                    pyo.Constraint(
                        tset,
                        side_comps,
                        rule=partial(rule_flow_mol_comp, props=props, port=port),
                    ),
                )
                if direction == "in":
                    setattr(
                        self,
                        f"{port_name}_mole_frac_eqn",
                        pyo.Constraint(
                            tset,
                            rule=partial(rule_mole_frac, port=port, comps=side_comps),
                        ),
                    )
                if direction == "out":
                    setattr(
                        self,
                        f"{port_name}_absent_comp_eqn",
                        pyo.Constraint(
                            tset,
                            absent_comp_list,
                            rule=partial(rule_absent_comp, props=props),
                        ),
                    )
                # Add a different port at the module level
                self.add_port(
                    name=port_name,
                    block=props,
                    doc=f"{side.capitalize()}-side {direction.capitalize()}let Port",
                )

        # This is net flow of power *into* the cell. In fuel cell mode, this will
        # be negative
        @self.Expression(tset)
        def electrical_work(b, t):
            return b.solid_oxide_cell.electrical_work[t] * b.number_cells

        self.potential_cell = pyo.Reference(self.solid_oxide_cell.potential)

        self.total_current = pyo.Var(
            tset,
            units=pyo.units.ampere,
            doc="Total current through module",
            initialize=0,
        )

        @self.Constraint(tset)
        def total_current_eqn(b, t):
            return (
                b.total_current[t]
                == b.solid_oxide_cell.total_current[t] * b.number_cells
            )

    def initialize_build(
        self,
        state_args_fuel=None,
        state_args_oxygen=None,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
        current_density_guess=None,
        temperature_guess=None,
    ):
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")

        solver_obj = get_solver(solver, optarg)

        number_cells_fixed = self.number_cells.fixed
        potential_cell_fixed = {}
        for t in self.flowsheet().time:
            potential_cell_fixed[t] = self.potential_cell[t].fixed
            self.potential_cell[t].fix()

        self.number_cells.fix()

        flags = {}
        for side, state_args in zip(
            ["fuel", "oxygen"],
            [state_args_fuel, state_args_oxygen],
        ):
            props_in = getattr(self, side + "_properties_in")
            inlet = getattr(self.solid_oxide_cell, side + "_inlet")
            comps = getattr(self.solid_oxide_cell, side + "_component_list")
            flags[side] = props_in.initialize(
                outlvl=outlvl,
                optarg=optarg,
                solver=solver,
                state_args=state_args,
                hold_state=True,
            )
            for t in self.flowsheet().time:
                inlet.temperature[t].value = pyo.value(props_in[t].temperature)
                inlet.pressure[t].value = pyo.value(props_in[t].pressure)
                inlet.flow_mol[t].value = pyo.value(
                    props_in[t].flow_mol / self.number_cells
                )
                for j in comps:
                    inlet.mole_frac_comp[t, j].value = pyo.value(
                        props_in[t].mole_frac_comp[j]
                    )

        self.solid_oxide_cell.initialize_build(
            outlvl=idaeslog.NOTSET,
            solver=solver,
            optarg=optarg,
            current_density_guess=current_density_guess,
            temperature_guess=temperature_guess,
        )

        # Without knowing the state variables of the property package, the best we can do is just initialize
        # the property blocks and then solve the entire block
        for side, state_args in zip(
            ["fuel", "oxygen"],
            [state_args_fuel, state_args_oxygen],
        ):
            props_out = getattr(self, side + "_properties_out")
            props_out.initialize(
                outlvl=outlvl,
                optarg=optarg,
                solver=solver,
                state_args=state_args,
            )
        common._init_solve_block(self, solver_obj, solve_log)

        init_log.info(f"{self.name} initialization complete.")

        self.fuel_properties_in.release_state(flags=flags["fuel"], outlvl=outlvl)
        self.oxygen_properties_in.release_state(flags=flags["oxygen"], outlvl=outlvl)
        if not number_cells_fixed:
            self.number_cells.unfix()
        for t in self.flowsheet().time:
            if not potential_cell_fixed[t]:
                self.potential_cell[t].unfix()

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

        gsf = iscale.get_scaling_factor

        def ssf(c, s):
            iscale.set_scaling_factor(c, s, overwrite=False)

        def cst(c, s):
            return iscale.constraint_scaling_transform(c, s, overwrite=False)

        s_num_cells = gsf(self.number_cells, default=1 / self.number_cells.value)

        for side in ["fuel", "oxygen"]:
            comps = getattr(self.solid_oxide_cell, f"{side}_component_list")
            channel = getattr(self.solid_oxide_cell, f"{side}_channel")
            for direction in ["in", "out"]:
                props = getattr(self, f"{side}_properties_{direction}")
                port_name = f"{side}_{direction}let"  # e.g., fuel_inlet, oxygen_outlet

                temperature_eqn = getattr(self, f"{port_name}_temperature_eqn")
                pressure_eqn = getattr(self, f"{port_name}_pressure_eqn")
                flow_mol_comp_eqn = getattr(self, f"{port_name}_flow_mol_comp_eqn")
                if direction == "out":
                    absent_comp_eqn = getattr(self, f"{port_name}_absent_comp_eqn")
                    absent_comp_list = getattr(self, f"absent_{side}_component_list")
                for t in self.flowsheet().time:
                    sT = gsf(props[t].temperature)
                    cst(temperature_eqn[t], sT)
                    sP = gsf(props[t].pressure)
                    cst(pressure_eqn[t], sP)
                    sflow = gsf(props[t].flow_mol)

                    if direction == "in":
                        # To avoid Reference issues, use channel variables directly
                        ssf(channel.temperature_inlet[t], sT)
                        ssf(channel.pressure_inlet[t], sP)
                        ssf(channel.flow_mol_inlet[t], sflow / s_num_cells)

                    for j in comps:
                        sx = gsf(props[t].mole_frac_comp[j])
                        cst(flow_mol_comp_eqn[t, j], sflow * sx)
                        if direction == "in":
                            ssf(channel.mole_frac_comp_inlet[t, j], sx)

                    if direction == "out":
                        for j in absent_comp_list:
                            sx = gsf(props[t].mole_frac_comp[j])
                            cst(absent_comp_eqn[t, j], sflow * sx)

        self.solid_oxide_cell.recursive_scaling()
        for t in self.flowsheet().time:
            sf_tot_current = gsf(self.solid_oxide_cell.total_current[t])
            ssf(self.total_current[t], sf_tot_current * s_num_cells)
            sf_tot_current = gsf(self.total_current[t])
            cst(self.total_current_eqn[t], sf_tot_current)

    def model_check(self):
        pass



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################

__author__ = "John Eslick"

import pytest
import idaes.models.properties.iapws95 as iapws95
from idaes.models.properties.tests.test_harness import PropertyTestHarness
from idaes.models.properties.general_helmholtz import helmholtz_available


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
class TestBasicMix(PropertyTestHarness):
    def configure(self):
        self.prop_pack = iapws95.Iapws95ParameterBlock
        self.param_args = {"phase_presentation": iapws95.PhaseType.MIX}
        self.prop_args = {}
        self.has_density_terms = True
        # Helmholtz package initialization has no solver calls, so can't fail
        self.skip_initialization_raises_exception_test = True


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
class TestBasicLV(PropertyTestHarness):
    def configure(self):
        self.prop_pack = iapws95.Iapws95ParameterBlock
        self.param_args = {"phase_presentation": iapws95.PhaseType.LG}
        self.prop_args = {}
        self.has_density_terms = True
        # Helmholtz package initialization has no solver calls, so can't fail
        self.skip_initialization_raises_exception_test = True


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
class TestBasicL(PropertyTestHarness):
    def configure(self):
        self.prop_pack = iapws95.Iapws95ParameterBlock
        self.param_args = {"phase_presentation": iapws95.PhaseType.L}
        self.prop_args = {}
        self.has_density_terms = True
        # Helmholtz package initialization has no solver calls, so can't fail
        self.skip_initialization_raises_exception_test = True


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
class TestBasicV(PropertyTestHarness):
    def configure(self):
        self.prop_pack = iapws95.Iapws95ParameterBlock
        self.param_args = {"phase_presentation": iapws95.PhaseType.G}
        self.prop_args = {}
        self.has_density_terms = True
        # Helmholtz package initialization has no solver calls, so can't fail
        self.skip_initialization_raises_exception_test = True



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
IDAES Moving Bed Model.

The moving bed model is a 1D axially discretized model with a gas and
solid phase and a counter-current flow direction. The model captures
the gas-solid interaction between both phases through reaction, mass
and heat transfer.

Equations written in this model were derived from:
(1) A. Ostace, A. Lee, C.O. Okoli, A.P. Burgard, D.C. Miller, D. Bhattacharyya,
Mathematical modeling of a moving-bed reactor for chemical looping combustion
of methane, in: M.R. Eden, M. Ierapetritou, G.P. Towler (Eds.),13th Int. Symp.
Process Syst. Eng. (PSE 2018), Computer-Aided Chemical Engineering 2018,
pp. 325–330 , San Diego, CA.
(2) C.O. Okoli, A. Ostace, S. Nadgouda, A. Lee, A. Tong, A.P. Burgard,
D. Bhattacharyya, D.C. Miller, 2020.
A framework for the optimization of chemical looping combustion processes,
Journal of Powder Technology, Vol. 365, pp 149 - 162.

Assumptions:
Property package contains temperature and pressure variables.
Property package contains minimum fluidization velocity.

"""

# Import Python libraries
import matplotlib.pyplot as plt

# Import Pyomo libraries
from pyomo.environ import (
    Var,
    Param,
    Reals,
    value,
    TransformationFactory,
    Constraint,
    check_optimal_termination,
    units as pyunits,
)
from pyomo.common.config import ConfigBlock, ConfigValue, In, Bool
from pyomo.util.calc_var_value import calculate_variable_from_constraint
from pyomo.dae import ContinuousSet

# Import IDAES cores
from idaes.core import (
    ControlVolume1DBlock,
    UnitModelBlockData,
    declare_process_block_class,
    MaterialBalanceType,
    EnergyBalanceType,
    MomentumBalanceType,
    FlowDirection,
    DistributedVars,
)
from idaes.core.util.config import (
    is_physical_parameter_block,
    is_reaction_parameter_block,
)
from idaes.core.util.exceptions import (
    ConfigurationError,
    BurntToast,
    InitializationError,
)
from idaes.core.util.tables import create_stream_table_dataframe
from idaes.core.util.constants import Constants as constants
from idaes.core.util.math import smooth_abs
from idaes.core.util.misc import add_object_reference
import idaes.logger as idaeslog
from idaes.core.util import scaling as iscale
from idaes.core.solvers import get_solver

__author__ = "Chinedu Okoli", "Anca Ostace"

# Set up logger
_log = idaeslog.getLogger(__name__)


@declare_process_block_class("MBR")
class MBRData(UnitModelBlockData):
    """Standard Moving Bed Unit Model Class."""

    # Create template for unit level config arguments
    CONFIG = UnitModelBlockData.CONFIG()

    # Unit level config arguments
    CONFIG.declare(
        "finite_elements",
        ConfigValue(
            default=10,
            domain=int,
            description="Number of finite elements length domain",
            doc="""Number of finite elements to use when discretizing length
domain (default=20)""",
        ),
    )
    CONFIG.declare(
        "length_domain_set",
        ConfigValue(
            default=[0.0, 1.0],
            domain=list,
            description="Number of finite elements length domain",
            doc="""length_domain_set - (optional) list of point to use to
initialize a new ContinuousSet if length_domain is not
provided (default = [0.0, 1.0])""",
        ),
    )
    CONFIG.declare(
        "transformation_method",
        ConfigValue(
            default="dae.finite_difference",
            description="Method to use for DAE transformation",
            doc="""Method to use to transform domain. Must be a method recognized
by the Pyomo TransformationFactory,
**default** - "dae.finite_difference".
**Valid values:** {
**"dae.finite_difference"** - Use a finite difference transformation method,
**"dae.collocation"** - use a collocation transformation method}""",
        ),
    )
    CONFIG.declare(
        "transformation_scheme",
        ConfigValue(
            default=None,
            domain=In([None, "BACKWARD", "FORWARD", "LAGRANGE-RADAU"]),
            description="Scheme to use for DAE transformation",
            doc="""Scheme to use when transforming domain. If specified,
this scheme is applied to discretizations in both the gas and solid phases.
In this case, ``gas_transformation_scheme`` and ``solid_transformation_scheme``
cannot be specified. See Pyomo documentation for supported schemes.
**default** - None.
**Valid values:** {
**None** - defaults to "BACKWARD" for finite difference transformation method,
and to "LAGRANGE-RADAU" for collocation transformation method,
**"BACKWARD"** - Use a finite difference transformation method,
**"FORWARD""** - use a finite difference transformation method,
**"LAGRANGE-RADAU""** - use a collocation transformation method}""",
        ),
    )
    CONFIG.declare(
        "gas_transformation_scheme",
        ConfigValue(
            default=None,
            domain=In([None, "BACKWARD", "FORWARD"]),
            description="Scheme to use for DAE transformation",
            doc="""Scheme to use when transforming length domain of the gas
phase. If this option is supplied, ``solid_transformation_scheme`` must be
supplied as well. This cannot be set (other than to ``None``) if the
``transformation_scheme`` option is set.
See Pyomo documentation for supported schemes.
**default** - None.
**Valid values:** {
**None** - defaults to the value of the ``transformation_scheme`` option,
**"BACKWARD"** - Use a finite difference transformation method,
**"FORWARD""** - use a finite difference transformation method}""",
        ),
    )
    CONFIG.declare(
        "solid_transformation_scheme",
        ConfigValue(
            default=None,
            domain=In([None, "BACKWARD", "FORWARD"]),
            description="Scheme to use for DAE transformation",
            doc="""Scheme to use when transforming length domain of the solid
phase. If this option is supplied, ``gas_transformation_scheme`` must be
supplied as well. This cannot be set (other than to ``None``) if the
``transformation_scheme`` option is set.
See Pyomo documentation for supported schemes,
**default** - None.
**Valid values:** {
**None** - defaults to the value of the ``transformation_scheme`` option,
**"BACKWARD"** - Use a finite difference transformation method,
**"FORWARD""** - use a finite difference transformation method}""",
        ),
    )
    CONFIG.declare(
        "collocation_points",
        ConfigValue(
            default=3,
            domain=int,
            description="Number of collocation points per finite element",
            doc="""Number of collocation points to use per finite element when
discretizing length domain (default=3)""",
        ),
    )
    CONFIG.declare(
        "flow_type",
        ConfigValue(
            default="counter_current",
            domain=In(["counter_current"]),
            description="Flow configuration of Moving Bed",
            doc="""Flow configuration of Moving Bed
- counter_current: gas side flows from 0 to 1
solid side flows from 1 to 0""",
        ),
    )
    CONFIG.declare(
        "material_balance_type",
        ConfigValue(
            default=MaterialBalanceType.componentTotal,
            domain=In(MaterialBalanceType),
            description="Material balance construction flag",
            doc="""Indicates what type of mass balance should be constructed,
**default** - MaterialBalanceType.componentTotal.
**Valid values:** {
**MaterialBalanceType.none** - exclude material balances,
**MaterialBalanceType.componentPhase** - use phase component balances,
**MaterialBalanceType.componentTotal** - use total component balances,
**MaterialBalanceType.elementTotal** - use total element balances,
**MaterialBalanceType.total** - use total material balance.}""",
        ),
    )
    CONFIG.declare(
        "energy_balance_type",
        ConfigValue(
            default=EnergyBalanceType.enthalpyTotal,
            domain=In(EnergyBalanceType),
            description="Energy balance construction flag",
            doc="""Indicates what type of energy balance should be constructed,
**default** - EnergyBalanceType.enthalpyTotal.
**Valid values:** {
**EnergyBalanceType.none** - exclude energy balances,
**EnergyBalanceType.enthalpyTotal** - single enthalpy balance for material,
**EnergyBalanceType.enthalpyPhase** - enthalpy balances for each phase,
**EnergyBalanceType.energyTotal** - single energy balance for material,
**EnergyBalanceType.energyPhase** - energy balances for each phase.}""",
        ),
    )
    CONFIG.declare(
        "momentum_balance_type",
        ConfigValue(
            default=MomentumBalanceType.pressureTotal,
            domain=In(MomentumBalanceType),
            description="Momentum balance construction flag",
            doc="""Indicates what type of momentum balance should be constructed,
**default** - MomentumBalanceType.pressureTotal.
**Valid values:** {
**MomentumBalanceType.none** - exclude momentum balances,
**MomentumBalanceType.pressureTotal** - single pressure balance for material,
**MomentumBalanceType.pressurePhase** - pressure balances for each phase,
**MomentumBalanceType.momentumTotal** - single momentum balance for material,
**MomentumBalanceType.momentumPhase** - momentum balances for each phase.}""",
        ),
    )
    CONFIG.declare(
        "has_pressure_change",
        ConfigValue(
            default=True,
            domain=Bool,
            description="Pressure change term construction flag",
            doc="""Indicates whether terms for pressure change should be
constructed,
**default** - False.
**Valid values:** {
**True** - include pressure change terms,
**False** - exclude pressure change terms.}""",
        ),
    )
    CONFIG.declare(
        "pressure_drop_type",
        ConfigValue(
            default="simple_correlation",
            domain=In(["simple_correlation", "ergun_correlation"]),
            description="Construction flag for type of pressure drop",
            doc="""Indicates what type of pressure drop correlation should be used,
**default** - "simple_correlation".
**Valid values:** {
**"simple_correlation"** - Use a simplified pressure drop correlation,
**"ergun_correlation"** - Use the Ergun equation.}""",
        ),
    )

    # Create template for phase specific config arguments
    _PhaseTemplate = UnitModelBlockData.CONFIG()
    _PhaseTemplate.declare(
        "has_equilibrium_reactions",
        ConfigValue(
            default=False,
            domain=Bool,
            description="Equilibrium reaction construction flag",
            doc="""Indicates whether terms for equilibrium controlled reactions
should be constructed,
**default** - True.
**Valid values:** {
**True** - include equilibrium reaction terms,
**False** - exclude equilibrium reaction terms.}""",
        ),
    )
    _PhaseTemplate.declare(
        "property_package",
        ConfigValue(
            default=None,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property calculations
(default = 'use_parent_value')
- 'use_parent_value' - get package from parent (default = None)
- a ParameterBlock object""",
        ),
    )
    _PhaseTemplate.declare(
        "property_package_args",
        ConfigValue(
            default={},
            domain=dict,
            description="Arguments for constructing gas property package",
            doc="""A dict of arguments to be passed to the PropertyBlockData
and used when constructing these
(default = 'use_parent_value')
- 'use_parent_value' - get package from parent (default = None)
- a dict (see property package for documentation)""",
        ),
    )
    _PhaseTemplate.declare(
        "reaction_package",
        ConfigValue(
            default=None,
            domain=is_reaction_parameter_block,
            description="Reaction package to use for control volume",
            doc="""Reaction parameter object used to define reaction calculations,
**default** - None.
**Valid values:** {
**None** - no reaction package,
**ReactionParameterBlock** - a ReactionParameterBlock object.}""",
        ),
    )
    _PhaseTemplate.declare(
        "reaction_package_args",
        ConfigBlock(
            implicit=True,
            implicit_domain=ConfigBlock,
            description="Arguments to use for constructing reaction packages",
            doc="""A ConfigBlock with arguments to be passed to a reaction block(s)
and used when constructing these,
**default** - None.
**Valid values:** {
see reaction package for documentation.}""",
        ),
    )

    # Create individual config blocks for gas and solid sides
    CONFIG.declare("gas_phase_config", _PhaseTemplate(doc="gas phase config arguments"))
    CONFIG.declare(
        "solid_phase_config", _PhaseTemplate(doc="solid phase config arguments")
    )

    # =========================================================================
    def build(self):
        """
        Begin building model (pre-DAE transformation).

        Args:
            None

        Returns:
            None
        """
        # Call UnitModel.build to build default attributes
        super().build()

        # Consistency check for transformation method and transformation scheme
        if (
            self.config.transformation_method == "dae.finite_difference"
            and self.config.transformation_scheme is None
        ):
            self.config.transformation_scheme = "BACKWARD"
        elif (
            self.config.transformation_method == "dae.collocation"
            and self.config.transformation_scheme is None
        ):
            self.config.transformation_scheme = "LAGRANGE-RADAU"
        elif (
            self.config.transformation_method == "dae.finite_difference"
            and self.config.transformation_scheme != "BACKWARD"
            and self.config.transformation_scheme != "FORWARD"
        ):
            raise ConfigurationError(
                "{} invalid value for "
                "transformation_scheme argument. "
                "Must be "
                "BACKWARD"
                " or "
                "FORWARD"
                " "
                "if transformation_method is"
                " "
                "dae.finite_difference"
                ".".format(self.name)
            )
        elif (
            self.config.transformation_method == "dae.collocation"
            and self.config.transformation_scheme != "LAGRANGE-RADAU"
        ):
            raise ConfigurationError(
                "{} invalid value for "
                "transformation_scheme argument."
                "Must be "
                "LAGRANGE-RADAU"
                " if "
                "transformation_method is"
                " "
                "dae.collocation"
                ".".format(self.name)
            )
        elif self.config.transformation_method != "dae.finite_difference" and (
            self.config.gas_transformation_scheme is not None
            or self.config.solid_transformation_scheme is not None
        ):
            raise ConfigurationError(
                "Invalid configuration of {}. transformation_method must be"
                ' "dae.finite_difference" if gas_transformation_scheme or'
                " solid_transformation_scheme are set".format(self.name)
            )
        elif (self.config.gas_transformation_scheme is None) != (
            self.config.solid_transformation_scheme is None
        ):
            raise ConfigurationError(
                "Invalid configuration of {}. Either both"
                " gas_transformation_scheme and solid_transformation_scheme"
                " must be set, or neither must be set. Got {} and {}.".format(
                    self.name,
                    self.config.gas_transformation_scheme,
                    self.config.solid_transformation_scheme,
                )
            )
        elif (self.config.transformation_scheme is not None) and (
            (self.config.gas_transformation_scheme is not None)
            or (self.config.solid_transformation_scheme is not None)
        ):
            raise ConfigurationError(
                "Invalid configuration of {}. transformation_scheme cannot be"
                " specified if gas_transformation_scheme and"
                " solid_transformation_scheme are specified.".format(self.name)
            )

        # Set flow directions for the control volume blocks
        # Gas flows from 0 to 1, solid flows from 1 to 0
        # An if statement is used here despite only one option to allow for
        # future extensions to other flow configurations
        if self.config.flow_type == "counter_current":
            set_direction_gas = FlowDirection.forward
            set_direction_solid = FlowDirection.backward
        else:
            raise BurntToast(
                "{} encountered unrecognized argument "
                "for flow type. Please contact the IDAES"
                " developers with this bug.".format(self.name)
            )

        # Set arguments for gas sides if homogeneous reaction block
        if self.config.gas_phase_config.reaction_package is not None:
            has_rate_reaction_gas_phase = True
        else:
            has_rate_reaction_gas_phase = False

        # Set arguments for gas and solid sides if heterogeneous reaction block
        if self.config.solid_phase_config.reaction_package is not None:
            has_rate_reaction_solid_phase = True
            has_mass_transfer_gas_phase = True
        else:
            has_rate_reaction_solid_phase = False
            has_mass_transfer_gas_phase = False

        # Set heat transfer terms
        if self.config.energy_balance_type != EnergyBalanceType.none:
            has_heat_transfer = True
        else:
            has_heat_transfer = False

        # Set heat of reaction terms
        if (
            self.config.energy_balance_type != EnergyBalanceType.none
            and self.config.gas_phase_config.reaction_package is not None
        ):
            has_heat_of_reaction_gas_phase = True
        else:
            has_heat_of_reaction_gas_phase = False

        if (
            self.config.energy_balance_type != EnergyBalanceType.none
            and self.config.solid_phase_config.reaction_package is not None
        ):
            has_heat_of_reaction_solid_phase = True
        else:
            has_heat_of_reaction_solid_phase = False

        # local aliases used to shorten object names
        solid_phase = self.config.solid_phase_config

        # Get units meta data from property packages
        units_meta_solid = solid_phase.property_package.get_metadata().get_derived_units

        # In this branch, we assume that either both gas and solid
        # transformation schemes are specified, or neither are specified.
        if (
            # solid and gas transformation schemes are specified
            (self.config.solid_transformation_scheme is not None)
            and (self.config.gas_transformation_scheme is not None)
        ):
            # Create length domain sets for gas and solid phases. This is
            # necessary so we can apply different discretizations to the
            # different phases.
            self._gas_tr_scheme = self.config.gas_transformation_scheme
            self._solid_tr_scheme = self.config.solid_transformation_scheme
            self.solid_length_domain = ContinuousSet(
                bounds=(0.0, 1.0),
                initialize=self.config.length_domain_set,
                doc="Solid phase normalized length domain",
            )
            self.gas_length_domain = ContinuousSet(
                bounds=(0.0, 1.0),
                initialize=self.config.length_domain_set,
                doc="Gas phase normalized length domain",
            )
            # Create an instance attribute "length_domain" for convenience.
            # Arbitrarily, use the solid phase length domain.
            #
            # We do not use Reference as it does not work for Sets.
            add_object_reference(self, "length_domain", self.solid_length_domain)
        else:
            # Neither gas nor solid transformation schemes are specified
            self._gas_tr_scheme = self.config.transformation_scheme
            self._solid_tr_scheme = self.config.transformation_scheme
            self.length_domain = ContinuousSet(
                bounds=(0.0, 1.0),
                initialize=self.config.length_domain_set,
                doc="Moving bed normalized length domain",
            )
            # Create instance attributes for gas and solid length domains
            add_object_reference(self, "solid_length_domain", self.length_domain)
            add_object_reference(self, "gas_length_domain", self.length_domain)

        self.bed_height = Var(
            domain=Reals,
            initialize=1,
            doc="Bed length",
            units=units_meta_solid("length"),
        )

        # =========================================================================
        # Build Control volume 1D for gas phase and populate gas control volume

        self.gas_phase = ControlVolume1DBlock(
            transformation_method=self.config.transformation_method,
            transformation_scheme=self._gas_tr_scheme,
            finite_elements=self.config.finite_elements,
            collocation_points=self.config.collocation_points,
            dynamic=self.config.dynamic,
            has_holdup=self.config.has_holdup,
            area_definition=DistributedVars.variant,
            property_package=self.config.gas_phase_config.property_package,
            property_package_args=self.config.gas_phase_config.property_package_args,
            reaction_package=self.config.gas_phase_config.reaction_package,
            reaction_package_args=self.config.gas_phase_config.reaction_package_args,
        )

        self.gas_phase.add_geometry(
            length_domain=self.gas_length_domain,
            length_domain_set=self.config.length_domain_set,
            length_var=self.bed_height,
            flow_direction=set_direction_gas,
        )

        self.gas_phase.add_state_blocks(
            information_flow=set_direction_gas, has_phase_equilibrium=False
        )

        if self.config.gas_phase_config.reaction_package is not None:
            self.gas_phase.add_reaction_blocks(
                has_equilibrium=self.config.gas_phase_config.has_equilibrium_reactions
            )

        self.gas_phase.add_material_balances(
            balance_type=self.config.material_balance_type,
            has_phase_equilibrium=False,
            has_mass_transfer=has_mass_transfer_gas_phase,
            has_rate_reactions=has_rate_reaction_gas_phase,
        )

        self.gas_phase.add_energy_balances(
            balance_type=self.config.energy_balance_type,
            has_heat_transfer=has_heat_transfer,
            has_heat_of_reaction=has_heat_of_reaction_gas_phase,
        )

        self.gas_phase.add_momentum_balances(
            balance_type=self.config.momentum_balance_type,
            has_pressure_change=self.config.has_pressure_change,
        )

        # =========================================================================
        # Build Control volume 1D for solid phase and populate solid control volume

        # Set argument for heterogeneous reaction block
        self.solid_phase = ControlVolume1DBlock(
            transformation_method=self.config.transformation_method,
            transformation_scheme=self._solid_tr_scheme,
            finite_elements=self.config.finite_elements,
            collocation_points=self.config.collocation_points,
            dynamic=self.config.dynamic,
            has_holdup=self.config.has_holdup,
            area_definition=DistributedVars.variant,
            property_package=self.config.solid_phase_config.property_package,
            property_package_args=self.config.solid_phase_config.property_package_args,
            reaction_package=self.config.solid_phase_config.reaction_package,
            reaction_package_args=self.config.solid_phase_config.reaction_package_args,
        )

        self.solid_phase.add_geometry(
            length_domain=self.solid_length_domain,
            length_domain_set=self.config.length_domain_set,
            length_var=self.bed_height,
            flow_direction=set_direction_solid,
        )

        self.solid_phase.add_state_blocks(
            information_flow=set_direction_solid, has_phase_equilibrium=False
        )

        if self.config.solid_phase_config.reaction_package is not None:
            # TODO - a generalization of the heterogeneous reaction block
            # The heterogeneous reaction block does not use the
            # add_reaction_blocks in control volumes as control volumes are
            # currently setup to handle only homogeneous reaction properties.
            # Thus appending the heterogeneous reaction block to the
            # solid state block is currently hard coded here.

            tmp_dict = dict(**self.config.solid_phase_config.reaction_package_args)
            tmp_dict["gas_state_block"] = self.gas_phase.properties
            tmp_dict["solid_state_block"] = self.solid_phase.properties
            tmp_dict["has_equilibrium"] = (
                self.config.solid_phase_config.has_equilibrium_reactions
            )
            tmp_dict["parameters"] = self.config.solid_phase_config.reaction_package
            self.solid_phase.reactions = (
                self.config.solid_phase_config.reaction_package.reaction_block_class(
                    self.flowsheet().time,
                    self.length_domain,
                    doc="Reaction properties in control volume",
                    **tmp_dict,
                )
            )

        self.solid_phase.add_material_balances(
            balance_type=self.config.material_balance_type,
            has_phase_equilibrium=False,
            has_mass_transfer=False,
            has_rate_reactions=has_rate_reaction_solid_phase,
        )

        self.solid_phase.add_energy_balances(
            balance_type=self.config.energy_balance_type,
            has_heat_transfer=has_heat_transfer,
            has_heat_of_reaction=has_heat_of_reaction_solid_phase,
        )

        self.solid_phase.add_momentum_balances(
            balance_type=MomentumBalanceType.none, has_pressure_change=False
        )

        # =========================================================================
        # Add ports
        # Add Ports for gas side
        self.add_inlet_port(name="gas_inlet", block=self.gas_phase)
        self.add_outlet_port(name="gas_outlet", block=self.gas_phase)

        # Add Ports for solid side
        self.add_inlet_port(name="solid_inlet", block=self.solid_phase)
        self.add_outlet_port(name="solid_outlet", block=self.solid_phase)

        # =========================================================================
        # Add performance equation method
        self._apply_transformation()
        self._make_performance()

    # =========================================================================
    def _apply_transformation(self):
        """
        Method to apply DAE transformation to the Control Volume length domain.
        Transformation applied will be based on the Control Volume
        configuration arguments.
        """
        if self.config.finite_elements is None:
            raise ConfigurationError(
                "{} was not provided a value for the finite_elements"
                " configuration argument. Please provide a valid value.".format(
                    self.name
                )
            )

        if self.config.transformation_method == "dae.finite_difference":
            self.discretizer = TransformationFactory(self.config.transformation_method)
            if (self.config.gas_transformation_scheme is not None) and (
                self.config.solid_transformation_scheme is not None
            ):
                self.discretizer.apply_to(
                    self,
                    wrt=self.gas_length_domain,
                    nfe=self.config.finite_elements,
                    scheme=self._gas_tr_scheme,
                )
                self.discretizer.apply_to(
                    self,
                    wrt=self.solid_length_domain,
                    nfe=self.config.finite_elements,
                    scheme=self._solid_tr_scheme,
                )
            else:
                self.discretizer.apply_to(
                    self,
                    wrt=self.length_domain,
                    nfe=self.config.finite_elements,
                    scheme=self.config.transformation_scheme,
                )
        elif self.config.transformation_method == "dae.collocation":
            self.discretizer = TransformationFactory(self.config.transformation_method)
            self.discretizer.apply_to(
                self,
                wrt=self.length_domain,
                nfe=self.config.finite_elements,
                ncp=self.config.collocation_points,
                scheme=self.config.transformation_scheme,
            )

    def _make_performance(self):
        """
        Constraints for unit model.

        Args:
            None

        Returns:
            None
        """
        # local aliases used to shorten object names
        gas_phase = self.config.gas_phase_config
        solid_phase = self.config.solid_phase_config

        # Get units meta data from property packages
        units_meta_gas = gas_phase.property_package.get_metadata().get_derived_units
        units_meta_solid = solid_phase.property_package.get_metadata().get_derived_units

        # Declare Mutable Parameters
        self.eps = Param(
            mutable=True,
            default=1e-8,
            doc="Smoothing Factor for Smooth IF Statements",
            units=pyunits.dimensionless,
        )

        # Unit Model variables
        self.bed_diameter = Var(
            initialize=1, doc="Reactor diameter", units=units_meta_solid("length")
        )
        self.bed_area = Var(
            domain=Reals,
            initialize=1,
            doc="Reactor cross-sectional area",
            units=units_meta_solid("area"),
        )

        # Phase specific variables
        self.velocity_superficial_gas = Var(
            self.flowsheet().time,
            self.length_domain,
            domain=Reals,
            initialize=0.05,
            doc="Gas superficial velocity",
            units=units_meta_gas("velocity"),
        )
        self.velocity_superficial_solid = Var(
            self.flowsheet().time,
            domain=Reals,
            initialize=0.005,
            doc="Solid superficial velocity",
            units=units_meta_solid("velocity"),
        )

        # Dimensionless numbers, mass and heat transfer coefficients
        self.Re_particle = Var(
            self.flowsheet().time,
            self.length_domain,
            domain=Reals,
            initialize=1.0,
            doc="Particle Reynolds number",
            units=pyunits.dimensionless,
        )

        self.Pr_particle = Var(
            self.flowsheet().time,
            self.length_domain,
            domain=Reals,
            initialize=1.0,
            doc="Prandtl number of gas in bed",
            units=pyunits.dimensionless,
        )

        self.Nu_particle = Var(
            self.flowsheet().time,
            self.length_domain,
            domain=Reals,
            initialize=1.0,
            doc="Particle Nusselt number",
            units=pyunits.dimensionless,
        )
        self.gas_solid_htc = Var(
            self.flowsheet().time,
            self.length_domain,
            domain=Reals,
            initialize=1.0,
            doc="Gas-solid heat transfer coefficient",
            units=units_meta_gas("heat_transfer_coefficient"),
        )

        # Fixed variables (these are parameters that can be estimated)
        self.bed_voidage = Var(
            domain=Reals, initialize=0.4, doc="Bed voidage", units=pyunits.dimensionless
        )
        self.bed_voidage.fix()

        # =========================================================================
        # Add performance equations

        # ---------------------------------------------------------------------
        # Geometry constraints

        # Bed area
        @self.Constraint(doc="Bed area")
        def bed_area_eqn(b):
            return b.bed_area == (constants.pi * (0.5 * b.bed_diameter) ** 2)

        # Area of gas side, and solid side
        @self.Constraint(self.flowsheet().time, self.length_domain, doc="Gas side area")
        def gas_phase_area(b, t, x):
            return (
                b.gas_phase.area[t, x]
                == pyunits.convert(b.bed_area, to_units=units_meta_gas("area"))
                * b.bed_voidage
            )

        @self.Constraint(
            self.flowsheet().time, self.length_domain, doc="Solid side area"
        )
        def solid_phase_area(b, t, x):
            return b.solid_phase.area[t, x] == b.bed_area * (1 - b.bed_voidage)

        # ---------------------------------------------------------------------
        # Hydrodynamic constraints

        # Gas superficial velocity
        @self.Constraint(
            self.flowsheet().time, self.length_domain, doc="Gas superficial velocity"
        )
        def gas_super_vel(b, t, x):
            return (
                b.velocity_superficial_gas[t, x]
                * pyunits.convert(b.bed_area, to_units=units_meta_gas("area"))
                * b.gas_phase.properties[t, x].dens_mol
                == b.gas_phase.properties[t, x].flow_mol
            )

        # Solid superficial velocity
        @self.Constraint(
            self.flowsheet().time, self.length_domain, doc="Solid superficial velocity"
        )
        # This equation uses inlet values to compute the constant solid
        # superficial velocity, and then computes the solid particle density
        # through the rest of the bed.
        def solid_super_vel(b, t, x):
            return (
                b.velocity_superficial_solid[t]
                * b.bed_area
                * b.solid_phase.properties[t, x].dens_mass_particle
                == b.solid_phase.properties[t, x].flow_mass
            )

        # Gas side pressure drop calculation
        if (
            self.config.has_pressure_change
            and self.config.pressure_drop_type == "simple_correlation"
        ):
            # Simplified pressure drop
            @self.Constraint(
                self.flowsheet().time,
                self.length_domain,
                doc="Gas side pressure drop calculation -" "simplified pressure drop",
            )
            def gas_phase_config_pressure_drop(b, t, x):
                #  0.2/s is a unitted constant in the correlation
                return pyunits.convert(
                    b.gas_phase.deltaP[t, x],
                    to_units=units_meta_solid("pressure") / units_meta_solid("length"),
                ) == -(0.2 / pyunits.s) * (
                    b.velocity_superficial_gas[t, x]
                    * (
                        pyunits.convert(
                            b.solid_phase.properties[t, x].dens_mass_particle,
                            to_units=units_meta_solid("density_mass"),
                        )
                        - b.gas_phase.properties[t, x].dens_mass
                    )
                )

        elif (
            self.config.has_pressure_change
            and self.config.pressure_drop_type == "ergun_correlation"
        ):
            # Ergun equation
            @self.Constraint(
                self.flowsheet().time,
                self.length_domain,
                doc="Gas side pressure drop calculation -" "Ergun equation",
            )
            def gas_phase_config_pressure_drop(b, t, x):
                return -pyunits.convert(
                    b.gas_phase.deltaP[t, x],
                    to_units=units_meta_solid("pressure") / units_meta_solid("length"),
                ) == (
                    (150 * pyunits.dimensionless)
                    * (1 - b.bed_voidage) ** 2
                    * b.gas_phase.properties[t, x].visc_d
                    * (
                        b.velocity_superficial_gas[t, x]
                        + pyunits.convert(
                            b.velocity_superficial_solid[t],
                            to_units=units_meta_solid("velocity"),
                        )
                    )
                    / (
                        pyunits.convert(
                            b.solid_phase.properties[t, x].params.particle_dia,
                            to_units=units_meta_solid("length"),
                        )
                        ** 2
                        * b.bed_voidage**3
                    )
                ) + (
                    (1.75 * pyunits.dimensionless)
                    * b.gas_phase.properties[t, x].dens_mass
                    * (1 - b.bed_voidage)
                    * (
                        b.velocity_superficial_gas[t, x]
                        + pyunits.convert(
                            b.velocity_superficial_solid[t],
                            to_units=units_meta_solid("velocity"),
                        )
                    )
                    ** 2
                    / (
                        pyunits.convert(
                            b.solid_phase.properties[t, x].params.particle_dia,
                            to_units=units_meta_solid("length"),
                        )
                        * b.bed_voidage**3
                    )
                )

            # to_units=deltaP_units)
            # The above expression has no absolute values - assumes:
            # (velocity_superficial_gas + velocity_superficial_solid) > 0

        elif self.config.has_pressure_change is False:
            pass

        else:
            raise BurntToast(
                "{} encountered unrecognized argument for "
                "the pressure drop correlation. Please contact the IDAES"
                " developers with this bug.".format(self.name)
            )
        # ---------------------------------------------------------------------
        # Reaction constraints

        # Build homogeneous reaction constraints
        if gas_phase.reaction_package is not None:
            # Gas side rate reaction extent
            @self.Constraint(
                self.flowsheet().time,
                self.length_domain,
                gas_phase.reaction_package.rate_reaction_idx,
                doc="Gas side rate reaction extent",
            )
            def gas_phase_config_rxn_ext(b, t, x, r):
                return b.gas_phase.rate_reaction_extent[t, x, r] == (
                    b.gas_phase.reactions[t, x].reaction_rate[r]
                    * b.gas_phase.area[t, x]
                )

        # Build heterogeneous reaction constraints
        if solid_phase.reaction_package is not None:
            # Solid side rate reaction extent
            @self.Constraint(
                self.flowsheet().time,
                self.length_domain,
                solid_phase.reaction_package.rate_reaction_idx,
                doc="Solid side rate reaction extent",
            )
            def solid_phase_config_rxn_ext(b, t, x, r):
                return b.solid_phase.rate_reaction_extent[t, x, r] == (
                    b.solid_phase.reactions[t, x].reaction_rate[r]
                    * b.solid_phase.area[t, x]
                )

            # Gas side heterogeneous rate reaction generation
            @self.Constraint(
                self.flowsheet().time,
                self.length_domain,
                gas_phase.property_package.phase_list,
                gas_phase.property_package.component_list,
                doc="Gas side heterogeneous" "rate reaction generation",
            )
            def gas_comp_hetero_rxn(b, t, x, p, j):
                return b.gas_phase.mass_transfer_term[t, x, p, j] == (
                    sum(
                        b.solid_phase.reactions[t, x].rate_reaction_stoichiometry[
                            r, p, j
                        ]
                        * b.solid_phase.reactions[t, x].reaction_rate[r]
                        for r in (solid_phase.reaction_package.rate_reaction_idx)
                    )
                    * b.solid_phase.area[t, x]
                )

        # ---------------------------------------------------------------------
        if self.config.energy_balance_type != EnergyBalanceType.none:
            # Solid phase - gas to solid heat transfer
            @self.Constraint(
                self.flowsheet().time,
                self.length_domain,
                doc="Solid phase - gas to solid heat transfer",
            )
            def solid_phase_heat_transfer(b, t, x):
                return pyunits.convert(
                    b.solid_phase.heat[t, x],
                    to_units=units_meta_gas("power") / units_meta_gas("length"),
                ) * pyunits.convert(
                    b.solid_phase.properties[t, x].params.particle_dia,
                    to_units=units_meta_gas("length"),
                ) == 6 * b.gas_solid_htc[
                    t, x
                ] * (
                    b.gas_phase.properties[t, x].temperature
                    - pyunits.convert(
                        b.solid_phase.properties[t, x].temperature,
                        to_units=units_meta_gas("temperature"),
                    )
                ) * pyunits.convert(
                    b.solid_phase.area[t, x], to_units=units_meta_gas("area")
                )

            # Dimensionless numbers, mass and heat transfer coefficients
            # Particle Reynolds number
            @self.Constraint(
                self.flowsheet().time,
                self.length_domain,
                doc="Particle Reynolds number",
            )
            def reynolds_number_particle(b, t, x):
                return (
                    b.Re_particle[t, x] * b.gas_phase.properties[t, x].visc_d
                    == b.velocity_superficial_gas[t, x]
                    * pyunits.convert(
                        b.solid_phase.properties[t, x].params.particle_dia,
                        to_units=units_meta_gas("length"),
                    )
                    * b.gas_phase.properties[t, x].dens_mass
                )

            # Prandtl number
            @self.Constraint(
                self.flowsheet().time,
                self.length_domain,
                doc="Prandtl number of gas in bed",
            )
            def prandtl_number(b, t, x):
                return (
                    b.Pr_particle[t, x] * b.gas_phase.properties[t, x].therm_cond
                    == b.gas_phase.properties[t, x].cp_mass
                    * b.gas_phase.properties[t, x].visc_d
                )

            # Particle Nusselt number
            @self.Constraint(
                self.flowsheet().time, self.length_domain, doc="Particle Nusselt number"
            )
            def nusselt_number_particle(b, t, x):
                return (
                    b.Nu_particle[t, x] ** 3
                    == (
                        2.0 + 1.1 * (smooth_abs(b.Re_particle[t, x], b.eps) ** 0.6) ** 3
                    )
                    * b.Pr_particle[t, x]
                )

            # Gas-solid heat transfer coefficient
            @self.Constraint(
                self.flowsheet().time,
                self.length_domain,
                doc="Gas-solid heat transfer coefficient",
            )
            def gas_solid_htc_eqn(b, t, x):
                return (
                    b.gas_solid_htc[t, x]
                    * pyunits.convert(
                        b.solid_phase.properties[t, x].params.particle_dia,
                        to_units=units_meta_gas("length"),
                    )
                    == b.Nu_particle[t, x] * b.gas_phase.properties[t, x].therm_cond
                )

            # Gas phase - gas to solid heat transfer
            @self.Constraint(
                self.flowsheet().time,
                self.length_domain,
                doc="Gas phase - gas to solid heat transfer",
            )
            def gas_phase_heat_transfer(b, t, x):
                return b.gas_phase.heat[t, x] * pyunits.convert(
                    b.solid_phase.properties[t, x].params.particle_dia,
                    to_units=units_meta_gas("length"),
                ) == -6 * b.gas_solid_htc[t, x] * (
                    b.gas_phase.properties[t, x].temperature
                    - pyunits.convert(
                        b.solid_phase.properties[t, x].temperature,
                        to_units=units_meta_gas("temperature"),
                    )
                ) * pyunits.convert(
                    b.solid_phase.area[t, x], to_units=units_meta_gas("area")
                )

        elif self.config.energy_balance_type == EnergyBalanceType.none:
            # If energy balance is none fix gas and solid temperatures to
            # solid inlet temperature
            @self.Constraint(
                self.flowsheet().time,
                self.length_domain,
                doc="Isothermal gas phase constraint",
            )
            def isothermal_gas_phase(b, t, x):
                if x == self.length_domain.first():
                    return Constraint.Skip
                else:
                    return (
                        b.gas_phase.properties[t, x].temperature
                        == b.solid_inlet.temperature[t]
                    )

            @self.Constraint(
                self.flowsheet().time,
                self.length_domain,
                doc="Isothermal solid phase constraint",
            )
            def isothermal_solid_phase(b, t, x):
                if x == self.length_domain.last():
                    return Constraint.Skip
                else:
                    return (
                        b.solid_phase.properties[t, x].temperature
                        == b.solid_inlet.temperature[t]
                    )

    # =========================================================================
    # Model initialization routine

    def initialize_build(
        blk,
        gas_phase_state_args=None,
        solid_phase_state_args=None,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
    ):
        """
        Initialization routine for MB unit.

        Keyword Arguments:
            gas_phase_state_args : a dict of arguments to be passed to the
                        property package(s) to provide an initial state for
                        initialization (see documentation of the specific
                        property package) (default = None).
            solid_phase_state_args : a dict of arguments to be passed to the
                        property package(s) to provide an initial state for
                        initialization (see documentation of the specific
                        property package) (default = None).
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None, use default solver)

        Returns:
            None
        """

        # Set up logger for initialization and solve
        init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(blk.name, outlvl, tag="unit")

        # Create solver
        opt = get_solver(solver, optarg)

        # ---------------------------------------------------------------------
        # local aliases used to shorten object names
        gas_phase = blk.config.gas_phase_config
        solid_phase = blk.config.solid_phase_config

        # Keep all unit model geometry constraints, derivative_var constraints,
        # and property block constraints active. Additionally, in control
        # volumes - keep conservation linking constraints and
        # holdup calculation (for dynamic flowsheets) constraints active

        geometry_constraints_terms = [
            "bed_area_eqn",
            "solid_phase_area",
            "gas_phase_area",
            "gas_phase_length",
            "solid_phase_length",
        ]
        endswith_terms = (
            "_disc_eq",
            "linking_constraint",
            "linking_constraints",
            "_holdup_calculation",
        )
        startswith_terms = "properties"

        for c in blk.component_objects(Constraint, descend_into=True):
            if (
                not c.parent_block().local_name.startswith(startswith_terms)
                and not c.local_name.endswith(endswith_terms)
                and c.local_name not in geometry_constraints_terms
            ):
                c.deactivate()

        # ---------------------------------------------------------------------
        # Initialize thermophysical property constraints
        init_log.info("Initialize Thermophysical Properties")
        # Initialize gas_phase block
        gas_phase_flags = blk.gas_phase.properties.initialize(
            state_args=gas_phase_state_args,
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            hold_state=True,
        )

        # Initialize solid_phase properties block
        solid_phase_flags = blk.solid_phase.properties.initialize(
            state_args=solid_phase_state_args,
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            hold_state=True,
        )

        init_log.info_high("Initialization Step 1 Complete.")

        # ---------------------------------------------------------------------
        # Initialize hydrodynamics (gas velocity)
        for t in blk.flowsheet().time:
            for x in blk.length_domain:
                calculate_variable_from_constraint(
                    blk.velocity_superficial_gas[t, x], blk.gas_super_vel[t, x]
                )

        blk.gas_super_vel.activate()

        init_log.info("Initialize Hydrodynamics")
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            results = opt.solve(blk, tee=slc.tee)
        if check_optimal_termination(results):
            init_log.info_high(
                "Initialization Step 2 {}.".format(idaeslog.condition(results))
            )
        else:
            _log.warning("{} Initialization Step 2 Failed.".format(blk.name))

        # ---------------------------------------------------------------------
        # Initialize mass balance - no reaction and no pressure drop

        # Unfix material balance state variables (including particle porosity)
        # but keep other states fixed
        blk.gas_phase.properties.release_state(gas_phase_flags)
        blk.solid_phase.properties.release_state(solid_phase_flags)
        for t in blk.flowsheet().time:
            for x in blk.length_domain:
                blk.gas_phase.properties[t, x].pressure.fix()
                blk.gas_phase.properties[t, x].temperature.fix()
                blk.solid_phase.properties[t, x].temperature.fix()

        blk.gas_phase.material_balances.activate()

        if gas_phase.reaction_package is not None:
            for t in blk.flowsheet().time:
                gas_rxn_gen = blk.gas_phase.rate_reaction_generation
                for x in blk.length_domain:
                    for p in gas_phase.property_package.phase_list:
                        for j in gas_phase.property_package.component_list:
                            (gas_rxn_gen[t, x, p, j].fix(0.0))

        blk.solid_phase.material_balances.activate()
        blk.solid_super_vel.activate()

        if solid_phase.reaction_package is not None:
            for t in blk.flowsheet().time:
                solid_rxn_gen = blk.solid_phase.rate_reaction_generation
                for x in blk.length_domain:
                    for p in solid_phase.property_package.phase_list:
                        for j in solid_phase.property_package.component_list:
                            (solid_rxn_gen[t, x, p, j].fix(0.0))

                # Gas side heterogeneous rate reaction generation
                for x in blk.length_domain:
                    for p in gas_phase.property_package.phase_list:
                        for j in gas_phase.property_package.component_list:
                            (blk.gas_phase.mass_transfer_term[t, x, p, j].fix(0.0))

        init_log.info("Initialize Mass Balances")
        init_log.info_high(
            "initialize mass balances - no reactions and no pressure drop"
        )
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            results = opt.solve(blk, tee=slc.tee)
        if check_optimal_termination(results):
            init_log.info_high(
                "Initialization Step 3a {}.".format(idaeslog.condition(results))
            )
        else:
            _log.warning("{} Initialization Step 3a Failed.".format(blk.name))

        # Initialize mass balance - with reaction and no pressure drop
        if gas_phase.reaction_package is not None:
            # local aliases used to shorten object names
            gas_rxn_gen = blk.gas_phase.rate_reaction_generation
            gas_phase_stoichiometry_eqn = (
                blk.gas_phase.rate_reaction_stoichiometry_constraint
            )

            # Initialize reaction property package
            blk.gas_phase.reactions.activate()
            for t in blk.flowsheet().time:
                for x in blk.length_domain:
                    obj = blk.gas_phase.reactions[t, x]
                    for c in obj.component_objects(Constraint, descend_into=False):
                        c.activate()

            blk.gas_phase.reactions.initialize(
                outlvl=outlvl, optarg=optarg, solver=solver
            )

            for t in blk.flowsheet().time:
                for x in blk.length_domain:
                    for r in gas_phase.reaction_package.rate_reaction_idx:
                        calculate_variable_from_constraint(
                            blk.gas_phase.rate_reaction_extent[t, x, r],
                            blk.gas_phase_config_rxn_ext[t, x, r],
                        )
                    for p in gas_phase.property_package.phase_list:
                        for j in gas_phase.property_package.component_list:
                            (gas_rxn_gen[t, x, p, j].unfix())
                            if not (
                                (
                                    blk.gas_phase.config.transformation_scheme
                                    != "FORWARD"
                                    and x == blk.length_domain.first()
                                )
                                or (
                                    blk.gas_phase.config.transformation_scheme
                                    == "FORWARD"
                                    and x == blk.length_domain.last()
                                )
                            ):
                                calculate_variable_from_constraint(
                                    gas_rxn_gen[t, x, p, j],
                                    gas_phase_stoichiometry_eqn[t, x, p, j],
                                )

            gas_phase_stoichiometry_eqn.activate()
            blk.gas_phase_config_rxn_ext.activate()

        if solid_phase.reaction_package is not None:
            # local aliases used to shorten object names
            solid_rxn_gen = blk.solid_phase.rate_reaction_generation
            solid_phase_stoichiometry_eqn = (
                blk.solid_phase.rate_reaction_stoichiometry_constraint
            )
            gas_mass_transfer_term = blk.gas_phase.mass_transfer_term

            # Initialize reaction property package
            blk.solid_phase.reactions.activate()
            for t in blk.flowsheet().time:
                for x in blk.length_domain:
                    obj = blk.solid_phase.reactions[t, x]
                    for c in obj.component_objects(Constraint, descend_into=False):
                        c.activate()

            blk.solid_phase.reactions.initialize(
                outlvl=outlvl, optarg=optarg, solver=solver
            )

            for t in blk.flowsheet().time:
                for x in blk.length_domain:
                    for p in gas_phase.property_package.phase_list:
                        for j in gas_phase.property_package.component_list:
                            (gas_mass_transfer_term[t, x, p, j].unfix())
                            calculate_variable_from_constraint(
                                gas_mass_transfer_term[t, x, p, j],
                                blk.gas_comp_hetero_rxn[t, x, p, j],
                            )
                for x in blk.length_domain:
                    for r in solid_phase.reaction_package.rate_reaction_idx:
                        calculate_variable_from_constraint(
                            blk.solid_phase.rate_reaction_extent[t, x, r],
                            blk.solid_phase_config_rxn_ext[t, x, r],
                        )
                    for p in solid_phase.property_package.phase_list:
                        for j in solid_phase.property_package.component_list:
                            (solid_rxn_gen[t, x, p, j].unfix())
                            if not (
                                (
                                    blk.solid_phase.config.transformation_scheme
                                    != "FORWARD"
                                    and x == blk.length_domain.first()
                                )
                                or (
                                    blk.solid_phase.config.transformation_scheme
                                    == "FORWARD"
                                    and x == blk.length_domain.last()
                                )
                            ):
                                calculate_variable_from_constraint(
                                    solid_rxn_gen[t, x, p, j],
                                    solid_phase_stoichiometry_eqn[t, x, p, j],
                                )

            blk.gas_comp_hetero_rxn.activate()
            blk.solid_phase.rate_reaction_stoichiometry_constraint.activate()
            blk.solid_phase_config_rxn_ext.activate()

        if (
            gas_phase.reaction_package is not None
            or solid_phase.reaction_package is not None
        ):
            init_log.info_high(
                "initialize mass balances - with reactions and no pressure drop"
            )
            with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
                results = opt.solve(blk, tee=slc.tee)
            if check_optimal_termination(results):
                init_log.info_high(
                    "Initialization Step 3b {}.".format(idaeslog.condition(results))
                )
            else:
                _log.warning("{} Initialization Step 3b Failed.".format(blk.name))

        # Initialize mass balance - with pressure drop
        for t in blk.flowsheet().time:
            for x in blk.length_domain:
                # Unfix all pressure variables except at the inlet
                if blk.gas_phase.properties[t, x].config.defined_state is False:
                    blk.gas_phase.properties[t, x].pressure.unfix()

        blk.gas_phase.pressure_balance.activate()

        if blk.config.has_pressure_change:
            blk.gas_phase_config_pressure_drop.activate()

            for t in blk.flowsheet().time:
                for x in blk.length_domain:
                    calculate_variable_from_constraint(
                        blk.gas_phase.deltaP[t, x],
                        blk.gas_phase_config_pressure_drop[t, x],
                    )

            init_log.info_high(
                "initialize mass balances - with reactions and pressure drop"
            )
            with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
                results = opt.solve(blk, tee=slc.tee)
            if check_optimal_termination(results):
                init_log.info_high(
                    "Initialization Step 3c {}.".format(idaeslog.condition(results))
                )
            else:
                _log.warning("{} Initialization Step 3c Failed.".format(blk.name))
        # ---------------------------------------------------------------------
        # Initialize energy balance
        if blk.config.energy_balance_type != EnergyBalanceType.none:
            # Initialize dimensionless numbers,
            # mass and heat transfer coefficients
            for t in blk.flowsheet().time:
                for x in blk.length_domain:
                    calculate_variable_from_constraint(
                        blk.Re_particle[t, x], blk.reynolds_number_particle[t, x]
                    )
                    calculate_variable_from_constraint(
                        blk.Pr_particle[t, x], blk.prandtl_number[t, x]
                    )
                    calculate_variable_from_constraint(
                        blk.Nu_particle[t, x], blk.nusselt_number_particle[t, x]
                    )
                    calculate_variable_from_constraint(
                        blk.gas_solid_htc[t, x], blk.gas_solid_htc_eqn[t, x]
                    )
                    calculate_variable_from_constraint(
                        blk.gas_phase.heat[t, x], blk.gas_phase_heat_transfer[t, x]
                    )
                    calculate_variable_from_constraint(
                        blk.solid_phase.heat[t, x], blk.solid_phase_heat_transfer[t, x]
                    )

            # Unfix temperatures
            for t in blk.flowsheet().time:
                for x in blk.length_domain:
                    # Unfix all gas temperature variables except at the inlet
                    if blk.gas_phase.properties[t, x].config.defined_state is False:
                        blk.gas_phase.properties[t, x].temperature.unfix()
                for x in blk.length_domain:
                    # Unfix all solid temperature variables except at the inlet
                    if blk.solid_phase.properties[t, x].config.defined_state is False:
                        blk.solid_phase.properties[t, x].temperature.unfix()

            blk.reynolds_number_particle.activate()
            blk.prandtl_number.activate()
            blk.nusselt_number_particle.activate()
            blk.gas_solid_htc_eqn.activate()

            # Activate gas phase energy balance equations
            blk.gas_phase_heat_transfer.activate()
            blk.gas_phase.enthalpy_balances.activate()

            # Activate solid phase energy balance equations
            blk.solid_phase_heat_transfer.activate()
            blk.solid_phase.enthalpy_balances.activate()

            init_log.info("Initialize Energy Balances")
            with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
                results = opt.solve(blk, tee=slc.tee)
            if check_optimal_termination(results):
                init_log.info_high(
                    "Initialization Step 4 {}.".format(idaeslog.condition(results))
                )
            else:
                raise InitializationError(
                    f"{blk.name} failed to initialize successfully. Please "
                    f"check the output logs for more information."
                )

        # Initialize energy balance
        if blk.config.energy_balance_type == EnergyBalanceType.none:
            for t in blk.flowsheet().time:
                for x in blk.length_domain:
                    # Unfix all gas temperature variables except at the inlet
                    if blk.gas_phase.properties[t, x].config.defined_state is False:
                        blk.gas_phase.properties[t, x].temperature.unfix()
                for x in blk.length_domain:
                    # Unfix all solid temperature variables except at the inlet
                    if blk.solid_phase.properties[t, x].config.defined_state is False:
                        blk.solid_phase.properties[t, x].temperature.unfix()

            blk.isothermal_gas_phase.activate()
            blk.isothermal_solid_phase.activate()

            init_log.info("Initialize Energy Balances")
            with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
                results = opt.solve(blk, tee=slc.tee)
            if check_optimal_termination(results):
                init_log.info_high(
                    "Initialization Step 4 {}.".format(idaeslog.condition(results))
                )
            else:
                raise InitializationError(
                    f"{blk.name} failed to initialize successfully. Please "
                    f"check the output logs for more information."
                )

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()
        # local aliases used to shorten object names
        gas_phase = self.config.gas_phase_config

        # Get units meta data from property packages
        units_meta_gas = gas_phase.property_package.get_metadata().get_derived_units

        # scale some variables
        if hasattr(self, "bed_diameter"):
            if iscale.get_scaling_factor(self.bed_diameter) is None:
                sf = 1 / value(self.bed_diameter)
                iscale.set_scaling_factor(self.bed_diameter, 1e-1 * sf)

        if hasattr(self, "bed_area"):
            if iscale.get_scaling_factor(self.bed_area) is None:
                sf = 1 / value(constants.pi * (0.5 * self.bed_diameter) ** 2)
                iscale.set_scaling_factor(self.bed_area, 1e-1 * sf)

        if hasattr(self.gas_phase, "area"):
            if iscale.get_scaling_factor(self.gas_phase.area) is None:
                sf = iscale.get_scaling_factor(self.bed_area)
                iscale.set_scaling_factor(self.gas_phase.area, sf)

        if hasattr(self.solid_phase, "area"):
            if iscale.get_scaling_factor(self.solid_phase.area) is None:
                sf = iscale.get_scaling_factor(self.bed_area)
                iscale.set_scaling_factor(self.solid_phase.area, sf)

        if self.config.energy_balance_type != EnergyBalanceType.none:
            if hasattr(self, "Re_particle"):
                for (t, x), v in self.Re_particle.items():
                    if iscale.get_scaling_factor(v) is None:
                        sf1 = iscale.get_scaling_factor(
                            self.gas_phase.properties[t, x].dens_mass
                        )
                        sf2 = 1 / value(
                            pyunits.convert(
                                self.solid_phase.properties[t, x].params.particle_dia,
                                to_units=units_meta_gas("length"),
                            )
                        )
                        iscale.set_scaling_factor(v, 1e-1 * sf1 * sf2)

            if hasattr(self, "Pr_particle"):
                for (t, x), v in self.Pr_particle.items():
                    if iscale.get_scaling_factor(v) is None:
                        sf1 = iscale.get_scaling_factor(
                            self.gas_phase.properties[t, x].cp_mass
                        )
                        sf2 = iscale.get_scaling_factor(
                            self.gas_phase.properties[t, x].visc_d
                        )
                        iscale.set_scaling_factor(v, sf1 * sf2)

            if hasattr(self, "Nu_particle"):
                for (t, x), v in self.Nu_particle.items():
                    if iscale.get_scaling_factor(v) is None:
                        sf = 1 / (
                            value(
                                (
                                    (
                                        2.0
                                        + 1.1
                                        * (
                                            smooth_abs(self.Re_particle[t, x], self.eps)
                                            ** 0.6
                                        )
                                        ** 3
                                    )
                                    * self.Pr_particle[t, x]
                                )
                                ** (1 / 3)
                            )
                        )
                        iscale.set_scaling_factor(v, 1e-1 * sf)

            if hasattr(self, "gas_solid_htc"):
                if iscale.get_scaling_factor(self.gas_solid_htc) is None:
                    for (t, x), v in self.gas_solid_htc.items():
                        sf1 = iscale.get_scaling_factor(self.Nu_particle[t, x])
                        sf2 = iscale.get_scaling_factor(
                            self.gas_phase.properties[t, x].therm_cond
                        )
                        iscale.set_scaling_factor(v, sf1 * sf2)

            # scale heat variables in CV1D
            for (t, x), v in self.gas_phase.heat.items():
                sf1 = iscale.get_scaling_factor(
                    self.gas_phase.properties[t, x].enth_mol
                )
                sf2 = iscale.get_scaling_factor(
                    self.gas_phase.properties[t, x].flow_mol
                )
                iscale.set_scaling_factor(v, sf1 * sf2)

            for (t, x), v in self.solid_phase.heat.items():
                sf1 = iscale.get_scaling_factor(
                    self.solid_phase.properties[t, x].enth_mass
                )
                sf2 = iscale.get_scaling_factor(
                    self.solid_phase.properties[t, x].flow_mass
                )
                iscale.set_scaling_factor(v, sf1 * sf2)

        # Scale some constraints
        if hasattr(self, "bed_area_eqn"):
            for c in self.bed_area_eqn.values():
                iscale.constraint_scaling_transform(
                    c, iscale.get_scaling_factor(self.bed_area), overwrite=False
                )

        if hasattr(self, "gas_phase_area"):
            for (t, x), c in self.gas_phase_area.items():
                iscale.constraint_scaling_transform(
                    c, iscale.get_scaling_factor(self.bed_area), overwrite=False
                )

        if hasattr(self, "solid_phase_area"):
            for (t, x), c in self.solid_phase_area.items():
                iscale.constraint_scaling_transform(
                    c, iscale.get_scaling_factor(self.bed_area), overwrite=False
                )

        if hasattr(self, "gas_super_vel"):
            for (t, x), c in self.gas_super_vel.items():
                iscale.constraint_scaling_transform(
                    c,
                    iscale.get_scaling_factor(self.gas_phase.properties[t, x].flow_mol),
                    overwrite=False,
                )

        if hasattr(self, "solid_super_vel"):
            for (t, x), c in self.solid_super_vel.items():
                iscale.constraint_scaling_transform(
                    c,
                    1e-1
                    * iscale.get_scaling_factor(
                        self.solid_phase.properties[t, x].flow_mass
                    ),
                    overwrite=False,
                )

        if hasattr(self, "gas_phase_config_pressure_drop"):
            for (t, x), c in self.gas_phase_config_pressure_drop.items():
                iscale.constraint_scaling_transform(
                    c,
                    iscale.get_scaling_factor(self.gas_phase.deltaP[t, x]),
                    overwrite=False,
                )

        if hasattr(self, "gas_phase_config_rxn_ext"):
            for (t, x, r), c in self.gas_phase_config_rxn_ext.items():
                sf1 = iscale.get_scaling_factor(
                    self.gas_phase.reactions[t, x].reaction_rate[r]
                )
                sf2 = iscale.get_scaling_factor(self.bed_area)
                iscale.constraint_scaling_transform(c, sf1 * sf2, overwrite=False)

        if hasattr(self, "solid_phase_config_rxn_ext"):
            for (t, x, r), c in self.solid_phase_config_rxn_ext.items():
                sf1 = iscale.get_scaling_factor(
                    self.solid_phase.reactions[t, x].reaction_rate[r]
                )
                sf2 = iscale.get_scaling_factor(self.bed_area)
                iscale.constraint_scaling_transform(c, sf1 * sf2, overwrite=False)

        if hasattr(self, "gas_comp_hetero_rxn"):
            for (t, x, p, j), c in self.gas_comp_hetero_rxn.items():
                iscale.constraint_scaling_transform(
                    c,
                    iscale.get_scaling_factor(
                        self.gas_phase.mass_transfer_term[t, x, p, j]
                    ),
                    overwrite=False,
                )

        if self.config.energy_balance_type != EnergyBalanceType.none:
            if hasattr(self, "reynolds_number_particle"):
                for (t, x), c in self.reynolds_number_particle.items():
                    iscale.constraint_scaling_transform(
                        c,
                        iscale.get_scaling_factor(self.Re_particle[t, x]),
                        overwrite=False,
                    )

            if hasattr(self, "prandtl_number"):
                for (t, x), c in self.prandtl_number.items():
                    iscale.constraint_scaling_transform(
                        c,
                        iscale.get_scaling_factor(self.Pr_particle[t, x]),
                        overwrite=False,
                    )

            if hasattr(self, "nusselt_number_particle"):
                for (t, x), c in self.nusselt_number_particle.items():
                    iscale.constraint_scaling_transform(
                        c,
                        iscale.get_scaling_factor(self.Nu_particle[t, x]),
                        overwrite=False,
                    )

            if hasattr(self, "gas_solid_htc_eqn"):
                for (t, x), c in self.gas_solid_htc_eqn.items():
                    iscale.constraint_scaling_transform(
                        c,
                        iscale.get_scaling_factor(self.gas_solid_htc[t, x]),
                        overwrite=False,
                    )

            if hasattr(self, "gas_phase_heat_transfer"):
                for (t, x), c in self.gas_phase_heat_transfer.items():
                    iscale.constraint_scaling_transform(
                        c,
                        iscale.get_scaling_factor(self.gas_phase.heat[t, x]),
                        overwrite=False,
                    )

            if hasattr(self, "solid_phase_heat_transfer"):
                for (t, x), c in self.solid_phase_heat_transfer.items():
                    iscale.constraint_scaling_transform(
                        c,
                        iscale.get_scaling_factor(self.solid_phase.heat[t, x]),
                        overwrite=False,
                    )

        elif self.config.energy_balance_type == EnergyBalanceType.none:
            if hasattr(self, "isothermal_gas_phase"):
                for (t, x), c in self.isothermal_gas_phase.items():
                    iscale.constraint_scaling_transform(
                        c,
                        iscale.get_scaling_factor(
                            self.gas_phase.properties[t, x].temperature
                        ),
                        overwrite=False,
                    )

            if hasattr(self, "isothermal_solid_phase"):
                for (t, x), c in self.isothermal_solid_phase.items():
                    iscale.constraint_scaling_transform(
                        c,
                        iscale.get_scaling_factor(
                            self.gas_phase.properties[t, x].temperature
                        ),
                        overwrite=False,
                    )

    def results_plot(blk):
        """
        Plot method for common moving bed variables

        Variables plotted:
            Tg : Temperature in gas phase
            Ts : Temperature in solid phase
            vg : Superficial gas velocity
            P : Pressure in gas phase
            Ftotal : Total molar flowrate of gas
            Mtotal : Total mass flowrate of solid
            Cg : Concentration of gas components in the gas phase
            y_frac : Mole fraction of gas components in the gas phase
            x_frac : Mass fraction of solid components in the solid phase
        """
        print()
        print(
            "================================= Reactor plots ==============="
            "=================="
        )
        # local aliases used to shorten object names
        gas_phase = blk.config.gas_phase_config
        solid_phase = blk.config.solid_phase_config

        Tg = []
        Ts = []
        P = []
        Ftotal = []
        Mtotal = []
        vg = []

        for t in blk.flowsheet().time:
            for x in blk.gas_phase.length_domain:
                vg.append(value(blk.velocity_superficial_gas[t, x]))

        fig_vg = plt.figure(1)
        plt.plot(blk.gas_phase.length_domain, vg, label="vg")
        plt.legend(loc=0, ncol=2)
        plt.grid()
        plt.xlabel("Bed height [-]")
        plt.ylabel("Superficial gas velocity [m/s]")
        fig_vg.savefig("superficial_vel.png")

        # Pressure profile
        for t in blk.flowsheet().time:
            for x in blk.gas_phase.length_domain:
                P.append(blk.gas_phase.properties[t, x].pressure.value)

        fig_P = plt.figure(2)
        plt.plot(blk.gas_phase.length_domain, P, label="P")
        plt.legend(loc=0, ncol=2)
        plt.grid()
        plt.xlabel("Bed height [-]")
        plt.ylabel("Total Pressure [Pa]")
        fig_P.savefig("Pressure.png")

        # Temperature profile
        for t in blk.flowsheet().time:
            for x in blk.gas_phase.length_domain:
                Tg.append(blk.gas_phase.properties[t, x].temperature.value)
            for x in blk.solid_phase.length_domain:
                Ts.append(blk.solid_phase.properties[t, x].temperature.value)
        fig_T = plt.figure(3)
        plt.plot(blk.gas_phase.length_domain, Tg, label="Tg")
        plt.plot(blk.solid_phase.length_domain, Ts, label="Ts")
        plt.legend(loc=0, ncol=2)
        plt.grid()
        plt.xlabel("Bed height [-]")
        plt.ylabel("Temperature [K]")
        fig_T.savefig("Temperature.png")

        # Bulk gas phase total molar flow rate
        for t in blk.flowsheet().time:
            for x in blk.gas_phase.length_domain:
                Ftotal.append(blk.gas_phase.properties[t, x].flow_mol.value)
        fig_Ftotal = plt.figure(4)
        plt.plot(blk.gas_phase.length_domain, Ftotal)
        plt.grid()
        plt.xlabel("Bed height [-]")
        plt.ylabel("Total molar gas flow rate [mol/s]")
        fig_Ftotal.savefig("Total_gas_flow.png")

        # Bulk solid phase total mass flow rate
        for t in blk.flowsheet().time:
            for x in blk.solid_phase.length_domain:
                Mtotal.append(blk.solid_phase.properties[t, x].flow_mass.value)
        fig_Mtotal = plt.figure(5)
        plt.plot(blk.solid_phase.length_domain, Mtotal)
        plt.grid()
        plt.xlabel("Bed height [-]")
        plt.ylabel("Solid total mass flow rate [kg/s]")
        fig_Mtotal.savefig("Total_solid_flow.png")

        # Gas phase mole fractions
        for t in blk.flowsheet().time:
            for j in gas_phase.property_package.component_list:
                y_frac = []
                for x in blk.gas_phase.length_domain:
                    y_frac.append(
                        value(blk.gas_phase.properties[t, x].mole_frac_comp[j])
                    )
                fig_y = plt.figure(6)
                plt.plot(blk.gas_phase.length_domain, y_frac, label=j)
        plt.legend(loc=0, ncol=len(gas_phase.property_package.component_list))
        plt.grid()
        plt.xlabel("Bed height [-]")
        plt.ylabel("y_frac [-]")
        fig_y.savefig("Gas_mole_fractions.png")

        # Solid phase mass fractions
        for t in blk.flowsheet().time:
            for j in solid_phase.property_package.component_list:
                x_frac = []
                for x in blk.solid_phase.length_domain:
                    x_frac.append(
                        value(blk.solid_phase.properties[t, x].mass_frac_comp[j])
                    )
                fig_x = plt.figure(7)
                plt.plot(blk.solid_phase.length_domain, x_frac, label=j)
        plt.legend(loc=0, ncol=len(solid_phase.property_package.component_list))
        plt.grid()
        plt.xlabel("Bed height [-]")
        plt.ylabel("x_frac [-]")
        fig_x.savefig("Solid_mass_fractions.png")

        # Gas phase concentrations
        for t in blk.flowsheet().time:
            for j in gas_phase.property_package.component_list:
                Cg = []
                for x in blk.gas_phase.length_domain:
                    Cg.append(blk.gas_phase.properties[t, x].dens_mol_comp[j].value)
                fig_Cg = plt.figure(8)
                plt.plot(blk.gas_phase.length_domain, Cg, label=j)
        plt.legend(loc=0, ncol=len(gas_phase.property_package.component_list))
        plt.grid()
        plt.xlabel("Bed height [-]")
        plt.ylabel("Concentration [mol/m3]")
        fig_Cg.savefig("Gas_concentration.png")

    def _get_stream_table_contents(self, time_point=0):
        return create_stream_table_dataframe(
            {
                "Gas Inlet": self.gas_inlet,
                "Gas Outlet": self.gas_outlet,
                "Solid Inlet": self.solid_inlet,
                "Solid Outlet": self.solid_outlet,
            },
            time_point=time_point,
        )
